# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Protocol

import grpc
from grpc import aio


class InterceptorFunc(Protocol):
    def amend_call_details(  # type: ignore[no-untyped-def]
        self, client_call_details, grpc_call_details_class: Any
    ) -> Any: ...


class SyncUnaryUnaryInterceptor(grpc.UnaryUnaryClientInterceptor):  # type: ignore[type-arg]
    def __init__(self, func: InterceptorFunc):
        self.func = func

    def intercept_unary_unary(self, continuation, client_call_details, request):  # type: ignore[no-untyped-def]
        new_details = self.func.amend_call_details(client_call_details, grpc.ClientCallDetails)
        return continuation(new_details, request)


class SyncUnaryStreamInterceptor(grpc.UnaryStreamClientInterceptor):  # type: ignore[type-arg]
    def __init__(self, func: InterceptorFunc):
        self.func = func

    def intercept_unary_stream(self, continuation, client_call_details, request):  # type: ignore[no-untyped-def]
        new_details = self.func.amend_call_details(client_call_details, grpc.ClientCallDetails)
        return continuation(new_details, request)


class SyncStreamUnaryInterceptor(grpc.StreamUnaryClientInterceptor):  # type: ignore[type-arg]
    def __init__(self, func: InterceptorFunc):
        self.func = func

    def intercept_stream_unary(  # type: ignore[no-untyped-def]
        self, continuation, client_call_details, request_iterator
    ):
        new_details = self.func.amend_call_details(client_call_details, grpc.ClientCallDetails)
        return continuation(new_details, request_iterator)


class SyncStreamStreamInterceptor(grpc.StreamStreamClientInterceptor):  # type: ignore[type-arg]
    def __init__(self, func: InterceptorFunc):
        self.func = func

    def intercept_stream_stream(  # type: ignore[no-untyped-def]
        self, continuation, client_call_details, request_iterator
    ):
        new_details = self.func.amend_call_details(client_call_details, grpc.ClientCallDetails)
        return continuation(new_details, request_iterator)


class AsyncUnaryUnaryInterceptor(aio.UnaryUnaryClientInterceptor):  # type: ignore[type-arg]
    def __init__(self, func: InterceptorFunc):
        self.func = func

    async def intercept_unary_unary(self, continuation, client_call_details, request):  # type: ignore[no-untyped-def]
        new_details = self.func.amend_call_details(client_call_details, aio.ClientCallDetails)
        return await continuation(new_details, request)


class AsyncUnaryStreamInterceptor(aio.UnaryStreamClientInterceptor):  # type: ignore[type-arg]
    def __init__(self, func: InterceptorFunc):
        self.func = func

    async def intercept_unary_stream(self, continuation, client_call_details, request):  # type: ignore[no-untyped-def]
        new_details = self.func.amend_call_details(client_call_details, aio.ClientCallDetails)
        return await continuation(new_details, request)


class AsyncStreamUnaryInterceptor(aio.StreamUnaryClientInterceptor):  # type: ignore[type-arg]
    def __init__(self, func: InterceptorFunc):
        self.func = func

    async def intercept_stream_unary(  # type: ignore[no-untyped-def]
        self, continuation, client_call_details, request_iterator
    ):
        new_details = self.func.amend_call_details(client_call_details, aio.ClientCallDetails)
        return await continuation(new_details, request_iterator)


class AsyncStreamStreamInterceptor(aio.StreamStreamClientInterceptor):  # type: ignore[type-arg]
    def __init__(self, func: InterceptorFunc):
        self.func = func

    async def intercept_stream_stream(  # type: ignore[no-untyped-def]
        self, continuation, client_call_details, request_iterator
    ):
        new_details = self.func.amend_call_details(client_call_details, aio.ClientCallDetails)
        return await continuation(new_details, request_iterator)



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from threading import Event
from typing import Callable, List

from grpc import ServicerContext

from buildgrid.server.logging import buildgrid_logger

LOGGER = buildgrid_logger(__name__)


class CancellationContext:
    def __init__(self, context: ServicerContext) -> None:
        """
        Creates a wrapper for a grpc.ServicerContext which allows determining if a gRPC request has been
        cancelled by the client. Callbacks may be added to this context which will be invoked if the
        underlying grpc.ServicerContext is triggered.
        """

        self._event = Event()
        self._context = context
        self._callbacks: List[Callable[[], None]] = []
        context.add_callback(self._on_callback)

    def is_cancelled(self) -> bool:
        return self._event.is_set()

    def _on_callback(self) -> None:
        LOGGER.debug("Request cancelled.")
        self._event.set()
        for callback in self._callbacks:
            callback()

    def on_cancel(self, callback: Callable[[], None]) -> None:
        self._callbacks.append(callback)
        if self.is_cancelled():
            callback()



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import time
from datetime import timedelta
from typing import Any, Callable, Dict, Iterator, Optional, TypeVar, cast

import grpc

from buildgrid.server.metrics_utils import publish_timer_metric

Func = TypeVar("Func", bound=Callable)  # type: ignore[type-arg]


def _service_metadata(*args: Any, _error: Optional[str] = None) -> Dict[str, str]:
    # If the decorator is being used for service methods, then we assume errors are already being handled.
    # If an error does get through somehow, assume this leads to an internal error.
    if len(args) == 3:
        if isinstance(context := args[2], grpc.ServicerContext):
            code = context.code()  # type: ignore[attr-defined]
            if code is None:
                code = grpc.StatusCode.OK if _error is None else grpc.StatusCode.INTERNAL
            return {"status": code.name}

    # In this case, the error handling is being used for a normal method. Report the error.
    if _error is not None:
        return {"exceptionType": _error}

    return {}


def timed(metric_name: str, **tags: str) -> Callable[[Func], Func]:
    def decorator(func: Func) -> Func:
        @functools.wraps(func)
        def return_wrapper(*args: Any, **kwargs: Any) -> Any:
            start_time = time.perf_counter()
            error: Optional[str] = None
            try:
                return func(*args, **kwargs)
            except Exception as e:
                error = e.__class__.__name__
                raise
            finally:
                run_time = timedelta(seconds=time.perf_counter() - start_time)
                publish_timer_metric(metric_name, run_time, **_service_metadata(*args, _error=error), **tags)

        @functools.wraps(func)
        def yield_wrapper(*args: Any, **kwargs: Any) -> Iterator[Any]:
            start_time = time.perf_counter()
            error: Optional[str] = None
            try:
                yield from func(*args, **kwargs)
            except Exception as e:
                error = e.__class__.__name__
                raise
            finally:
                run_time = timedelta(seconds=time.perf_counter() - start_time)
                publish_timer_metric(metric_name, run_time, **_service_metadata(*args, _error=error), **tags)

        if inspect.isgeneratorfunction(func):
            return cast(Func, yield_wrapper)
        return cast(Func, return_wrapper)

    return decorator



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add a table for request metadata

Revision ID: e83194af8292
Revises: 316ad77858af
Create Date: 2024-06-14 12:49:52.934372

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e83194af8292"
down_revision = "316ad77858af"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "request_metadata",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("tool_name", sa.String(), nullable=True),
        sa.Column("tool_version", sa.String(), nullable=True),
        sa.Column("invocation_id", sa.String(), nullable=True),
        sa.Column("correlated_invocations_id", sa.String(), nullable=True),
        sa.Column("action_mnemonic", sa.String(), nullable=True),
        sa.Column("target_id", sa.String(), nullable=True),
        sa.Column("configuration_id", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "tool_name",
            "tool_version",
            "invocation_id",
            "correlated_invocations_id",
            "action_mnemonic",
            "target_id",
            "configuration_id",
            name="unique_metadata_constraint",
        ),
    )

    # We need to use `batch_alter_table` for SQLite, where `ALTER TABLE` isn't available
    with op.batch_alter_table("operations") as batch_op:
        batch_op.add_column(sa.Column("request_metadata_id", sa.Integer(), nullable=True))
        batch_op.create_foreign_key(
            "fk_operation_request_metadata", "request_metadata", ["request_metadata_id"], ["id"]
        )


def downgrade() -> None:
    with op.batch_alter_table("operations") as batch_op:
        batch_op.drop_constraint("fk_operation_request_metadata", type_="foreignkey")
        batch_op.drop_column("request_metadata_id")

    op.drop_table("request_metadata")



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
OperationsInstance
==================
An instance of the LongRunningOperations Service.
"""

from typing import Optional, Tuple

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import RequestMetadata
from buildgrid._protos.buildgrid.v2.identity_pb2 import ClientIdentity
from buildgrid._protos.google.longrunning.operations_pb2 import DESCRIPTOR as OPS_DESCRIPTOR
from buildgrid._protos.google.longrunning.operations_pb2 import ListOperationsResponse, Operation
from buildgrid.server.exceptions import InvalidArgumentError, NotFoundError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.operations.filtering import DEFAULT_OPERATION_FILTERS, FilterParser
from buildgrid.server.scheduler import Scheduler
from buildgrid.server.servicer import Instance
from buildgrid.server.settings import DEFAULT_MAX_LIST_OPERATION_PAGE_SIZE

LOGGER = buildgrid_logger(__name__)


class OperationsInstance(Instance):
    SERVICE_NAME = OPS_DESCRIPTOR.services_by_name["Operations"].full_name

    def __init__(
        self, scheduler: Scheduler, max_list_operations_page_size: int = DEFAULT_MAX_LIST_OPERATION_PAGE_SIZE
    ) -> None:
        self.scheduler = scheduler
        self._max_list_operations_page_size = max_list_operations_page_size

    # --- Public API ---

    def start(self) -> None:
        self.scheduler.start()

    def stop(self) -> None:
        self.scheduler.stop()
        LOGGER.info("Stopped Operations.")

    def get_operation(
        self, operation_name: str
    ) -> Tuple[Operation, Optional[RequestMetadata], Optional[ClientIdentity]]:
        try:
            operation = self.scheduler.load_operation(operation_name)

        except NotFoundError:
            raise InvalidArgumentError(f"Operation name does not exist: [{operation_name}]")

        metadata = self.scheduler.get_operation_request_metadata_by_name(operation_name)
        client_identity = self.scheduler.get_client_identity_by_operation(operation_name)
        return operation, metadata, client_identity

    def list_operations(
        self, filter_string: str, page_size: Optional[int], page_token: Optional[str]
    ) -> ListOperationsResponse:
        if page_size and page_size > self._max_list_operations_page_size:
            raise InvalidArgumentError(f"The maximum page size is {self._max_list_operations_page_size}.")
        if not page_size:
            page_size = self._max_list_operations_page_size

        operation_filters = FilterParser.parse_listoperations_filters(filter_string)
        if not operation_filters:
            operation_filters = DEFAULT_OPERATION_FILTERS

        response = ListOperationsResponse()

        results, next_token = self.scheduler.list_operations(operation_filters, page_size, page_token)
        response.operations.extend(results)
        response.next_page_token = next_token

        return response

    def delete_operation(self, job_name: str) -> None:
        """DeleteOperation is not supported in BuildGrid."""
        pass

    def cancel_operation(self, operation_name: str) -> None:
        try:
            self.scheduler.cancel_operation(operation_name)

        except NotFoundError:
            raise InvalidArgumentError(f"Operation name does not exist: [{operation_name}]")



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import re
from itertools import product
from threading import Event
from typing import Any, Iterator, List, Set, Tuple

from buildgrid.server.cleanup.janitor.config import JanitorConfig
from buildgrid.server.cleanup.janitor.index import IndexLookup
from buildgrid.server.cleanup.janitor.utils import check_bucket_versioning, get_s3_client
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.threading import ContextWorker

LOGGER = buildgrid_logger(__name__)


class S3Janitor:

    def __init__(self, config: JanitorConfig, index: IndexLookup):
        self._bucket_regex = re.compile(config.s3.bucket_regex)
        self._index = index
        self._path_prefix = config.s3.path_prefix
        self._s3 = get_s3_client(config.s3)
        self._sleep_interval = config.sleep_interval
        self._hash_prefix_size = config.s3.hash_prefix_size

        self._stop_requested = Event()
        self._worker = ContextWorker(target=self.run, name="Janitor", on_shutdown_requested=self._stop_requested.set)

    def enumerate_versioned_bucket(self, bucket: str, prefix: str) -> Iterator[Set[Tuple[str, str]]]:
        pages = self._s3.get_paginator("list_object_versions").paginate(Bucket=bucket, Prefix=prefix)
        for page in pages:
            if "Versions" not in page:
                continue

            digest_version_pairs = {(item["Key"], item["VersionId"]) for item in page["Versions"]}
            yield digest_version_pairs

    def enumerate_unversioned_bucket(self, bucket: str, prefix: str) -> Iterator[Set[Tuple[str, str]]]:
        pages = self._s3.get_paginator("list_objects").paginate(Bucket=bucket, Prefix=prefix)
        for page in pages:
            if "Contents" not in page:
                continue

            digest_version_pairs = {(item["Key"], "") for item in page["Contents"]}
            yield digest_version_pairs

    def delete_s3_entries(self, bucket: str, digest_versions: Set[Tuple[str, str]]) -> List[str]:
        LOGGER.info("Deleting orphaned blobs from S3.", tags=dict(digest_count=len(digest_versions)))
        response = self._s3.delete_objects(
            Bucket=bucket,
            Delete={
                "Objects": [{"Key": digest, "VersionId": version} for digest, version in digest_versions],
                "Quiet": False,
            },
        )
        return [key["Key"] for key in response.get("Deleted", [])]

    def get_buckets(self) -> List[str]:
        response = self._s3.list_buckets()
        return [
            bucket["Name"] for bucket in response["Buckets"] if self._bucket_regex.search(bucket["Name"]) is not None
        ]

    # Generate all the hash prefixes and shuffle them to reduce the likelihood of
    # two janitors cleaning the same hash prefix
    def generate_prefixes(self) -> List[str]:
        if self._hash_prefix_size:
            prefixes = [
                (self._path_prefix + "/" if self._path_prefix else "") + "".join(x)
                for x in product("0123456789abcdef", repeat=self._hash_prefix_size)
            ]
            random.shuffle(prefixes)
        else:
            prefixes = [self._path_prefix]
        return prefixes

    def cleanup_bucket(self, bucket: str) -> int:
        LOGGER.info("Cleaning up bucket.", tags=dict(bucket=bucket))

        deleted_count = 0
        if check_bucket_versioning(self._s3, bucket):
            enumeration = self.enumerate_versioned_bucket
        else:
            enumeration = self.enumerate_unversioned_bucket

        for prefix in self.generate_prefixes():
            deleted_count_for_prefix = 0
            for page in enumeration(bucket, prefix):
                # Create a mapping between a digest as stored in S3 and a digest as stored in the index
                # by stripping off any prefix and removing all '/' used by hash_prefix_size
                digest_map = {digest: digest.replace(self._path_prefix, "").replace("/", "") for digest, _ in page}

                missing_digest_versions = set(
                    digest_version
                    for digest_version in page
                    if digest_map[digest_version[0]] in self._index.get_missing_digests(set(digest_map.values()))
                )
                if missing_digest_versions:
                    self.delete_s3_entries(bucket, missing_digest_versions)
                    deleted_count_for_prefix += len(missing_digest_versions)
            LOGGER.info(
                "Deleted blobs from bucket prefix.",
                tags=dict(digest_count=deleted_count_for_prefix, bucket=bucket, prefix=prefix),
            )
            deleted_count += deleted_count_for_prefix

        LOGGER.info("Deleted blobs total from bucket.", tags=dict(digest_count=deleted_count, bucket=bucket))
        return deleted_count

    def start(self) -> None:
        self._worker.start()
        self._worker.wait()

    def stop(self, *args: Any, **kwargs: Any) -> None:
        self._worker.stop()

    def run(self, stop_requested: Event) -> None:
        random.seed()
        while not stop_requested.is_set():
            bucket_names = self.get_buckets()

            # Shuffle the bucket names to reduce the likelihood of two janitors
            # concurrently cleaning the same bucket.
            random.shuffle(bucket_names)

            for bucket in bucket_names:
                self.cleanup_bucket(bucket)

            stop_requested.wait(timeout=self._sleep_interval)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright (C) 2024 Bloomberg LP

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 <http://www.apache.org/licenses/LICENSE-2.0>

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class ClientIdentity(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ACTOR_FIELD_NUMBER: builtins.int
    SUBJECT_FIELD_NUMBER: builtins.int
    WORKFLOW_FIELD_NUMBER: builtins.int
    actor: builtins.str
    subject: builtins.str
    workflow: builtins.str
    def __init__(
        self,
        *,
        actor: builtins.str = ...,
        subject: builtins.str = ...,
        workflow: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["actor", b"actor", "subject", b"subject", "workflow", b"workflow"]) -> None: ...

global___ClientIdentity = ClientIdentity



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""partial_indices

Revision ID: 38340cc0cfb7
Revises: 1553978c6e69
Create Date: 2020-10-20 08:56:42.559227

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "38340cc0cfb7"
down_revision = "1553978c6e69"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_context().bind
    assert bind is not None, "Alembic database connection is missing"

    if bind.dialect.name == "postgresql":
        op.create_index(
            "ix_worker_start_timestamp",
            "jobs",
            ["worker_start_timestamp"],
            unique=False,
            postgresql_where=sa.text("worker_start_timestamp IS NOT NULL"),
        )
        op.create_index(
            "ix_worker_completed_timestamp",
            "jobs",
            ["worker_completed_timestamp"],
            unique=False,
            postgresql_where=sa.text("worker_completed_timestamp IS NOT NULL"),
        )
    elif bind.dialect.name == "sqlite":
        op.create_index(
            "ix_worker_start_timestamp",
            "jobs",
            ["worker_start_timestamp"],
            unique=False,
            sqlite_where=sa.text("worker_start_timestamp IS NOT NULL"),
        )
        op.create_index(
            "ix_worker_completed_timestamp",
            "jobs",
            ["worker_completed_timestamp"],
            unique=False,
            sqlite_where=sa.text("worker_completed_timestamp IS NOT NULL"),
        )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_worker_start_timestamp", table_name="jobs")
    op.drop_index("ix_worker_completed_timestamp", table_name="jobs")
    # ### end Alembic commands ###



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add bots table

Revision ID: fcb6e8f09a1d
Revises: b57c30a687fa
Create Date: 2023-06-07 11:10:58.114752

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "fcb6e8f09a1d"
down_revision = "b57c30a687fa"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "bots",
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("bot_id", sa.String(), nullable=False),
        sa.Column("last_update_timestamp", sa.DateTime(), nullable=False),
        sa.Column("bot_status", sa.Integer(), nullable=False),
        sa.Column("lease_id", sa.String(), nullable=True),
        sa.Column("instance_name", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("name"),
    )
    op.create_index(op.f("ix_bots_bot_id"), "bots", ["bot_id"], unique=False)
    op.create_index(op.f("ix_bots_last_update_timestamp"), "bots", ["last_update_timestamp"], unique=False)
    op.create_index(op.f("ix_bots_name"), "bots", ["name"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_bots_name"), table_name="bots")
    op.drop_index(op.f("ix_bots_last_update_timestamp"), table_name="bots")
    op.drop_index(op.f("ix_bots_bot_id"), table_name="bots")
    op.drop_table("bots")
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: buildgrid/v2/identity.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x62uildgrid/v2/identity.proto\x12\x0c\x62uildgrid.v2\"B\n\x0e\x43lientIdentity\x12\r\n\x05\x61\x63tor\x18\x01 \x01(\t\x12\x0f\n\x07subject\x18\x02 \x01(\t\x12\x10\n\x08workflow\x18\x03 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'buildgrid._protos.buildgrid.v2.identity_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_CLIENTIDENTITY']._serialized_start=45
  _globals['_CLIENTIDENTITY']._serialized_end=111
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextvars
import threading
from concurrent import futures
from concurrent.futures import Future
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Optional, Tuple, TypeVar

from buildgrid.server.logging import buildgrid_logger

_T = TypeVar("_T")

if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    _P = ParamSpec("_P")


LOGGER = buildgrid_logger(__name__)


class ContextThreadPoolExecutor(futures.ThreadPoolExecutor):
    def __init__(
        self,
        max_workers: Optional[int] = None,
        thread_name_prefix: str = "",
        initializer: Optional[Callable[[], Any]] = None,
        initargs: "Tuple[Any, ...]" = (),
        immediate_copy: bool = False,
    ) -> None:
        """
        Create a thread pool executor which forwards context from the creating thread.

        immediate_copy if true, copies the context when this threadpool object is created.
        If false, the context will be copied as jobs are submitted to it.
        """

        self._init_ctx: Optional[contextvars.Context] = None
        if immediate_copy:
            self._init_ctx = contextvars.copy_context()

        super().__init__(
            max_workers=max_workers,
            thread_name_prefix=thread_name_prefix,
            initializer=initializer,
            initargs=initargs,
        )

    def submit(self, fn: "Callable[_P, _T]", *args: "_P.args", **kwargs: "_P.kwargs") -> "Future[_T]":
        if self._init_ctx is None:
            run = contextvars.copy_context().run
        else:
            run = self._init_ctx.copy().run

        # In newer versions of grpcio (>=1.60.0), a context is used, but is not copied from
        # the initializer thread. We can use our own instead.
        if isinstance(getattr(fn, "__self__", None), contextvars.Context):
            return super().submit(run, *args, **kwargs)  # type: ignore[arg-type]
        return super().submit(run, fn, *args, **kwargs)  # type: ignore[arg-type]


class ContextThread(threading.Thread):
    def __init__(
        self,
        target: "Callable[_P, _T]",
        name: Optional[str] = None,
        args: Iterable[Any] = (),
        kwargs: Optional[Dict[str, Any]] = None,
        *,
        daemon: Optional[bool] = None,
    ) -> None:
        ctx = contextvars.copy_context()
        super().__init__(
            target=ctx.copy().run,
            name=name,
            args=(target, *args),
            kwargs=kwargs,
            daemon=daemon,
        )


class ContextWorker:
    def __init__(
        self,
        target: Callable[[threading.Event], None],
        name: Optional[str] = None,
        *,
        on_shutdown_requested: Optional[Callable[[], None]] = None,
    ) -> None:
        """
        Run a long-lived task in a thread, where the method is provided an Event that indicates if
        shutdown is requested. We delay creating the thread until started to allow the context
        to continue to be populated.
        """

        self._shutdown_requested = threading.Event()
        self._thread: Optional[ContextThread] = None

        self._target = target
        self._name = name
        self._on_shutdown_requested = on_shutdown_requested

    def is_alive(self) -> bool:
        return self._thread is not None and self._thread.is_alive()

    def __enter__(self) -> "ContextWorker":
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.stop()

    def start(self) -> None:
        if not self._thread:
            self._thread = ContextThread(
                target=lambda: self._target(self._shutdown_requested), name=self._name, daemon=True
            )
            self._thread.start()

    def stop(self) -> None:
        if not self._shutdown_requested.is_set():
            LOGGER.info("Stopping worker.", tags=dict(name=self._name))
            self._shutdown_requested.set()
            if self._on_shutdown_requested:
                self._on_shutdown_requested()
            if self._thread:
                self._thread.join()
            LOGGER.info("Stopped worker.", tags=dict(name=self._name))

    def wait(self, timeout: Optional[float] = None) -> None:
        if self._thread:
            self._thread.join(timeout=timeout)



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""New database with all earlier migrations squashed

Revision ID: e287a533bbd7
Revises:
Create Date: 2020-02-06 13:36:00.607572

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e287a533bbd7"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic ###
    op.create_table(
        "index",
        sa.Column("digest_hash", sa.String(), nullable=False),
        sa.Column("digest_size_bytes", sa.Integer(), nullable=False),
        sa.Column("accessed_timestamp", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("digest_hash"),
    )
    op.create_index(op.f("ix_index_accessed_timestamp"), "index", ["accessed_timestamp"], unique=False)
    op.create_index(op.f("ix_index_digest_hash"), "index", ["digest_hash"], unique=False)
    op.create_table(
        "jobs",
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("action_digest", sa.String(), nullable=False),
        sa.Column("priority", sa.Integer(), nullable=False),
        sa.Column("stage", sa.Integer(), nullable=False),
        sa.Column("do_not_cache", sa.Boolean(), nullable=False),
        sa.Column("cancelled", sa.Boolean(), nullable=False),
        sa.Column("queued_timestamp", sa.DateTime(), nullable=True),
        sa.Column("queued_time_duration", sa.Integer(), nullable=True),
        sa.Column("worker_start_timestamp", sa.DateTime(), nullable=True),
        sa.Column("worker_completed_timestamp", sa.DateTime(), nullable=True),
        sa.Column("result", sa.String(), nullable=True),
        sa.Column("assigned", sa.Boolean(), nullable=True),
        sa.Column("n_tries", sa.Integer(), nullable=True),
        sa.Column("platform_requirements", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("name"),
    )
    op.create_index(op.f("ix_jobs_action_digest"), "jobs", ["action_digest"], unique=False)
    op.create_index(op.f("ix_jobs_priority"), "jobs", ["priority"], unique=False)
    op.create_index(op.f("ix_jobs_stage"), "jobs", ["stage"], unique=False)
    op.create_table(
        "leases",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("job_name", sa.String(), nullable=False),
        sa.Column("status", sa.Integer(), nullable=True),
        sa.Column("state", sa.Integer(), nullable=False),
        sa.Column("worker_name", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(["job_name"], ["jobs.name"], onupdate="CASCADE", ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_leases_job_name"), "leases", ["job_name"], unique=False)
    op.create_index(op.f("ix_leases_worker_name"), "leases", ["worker_name"], unique=False)
    op.create_table(
        "operations",
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("job_name", sa.String(), nullable=False),
        sa.Column("cancelled", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(["job_name"], ["jobs.name"], onupdate="CASCADE", ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("name"),
    )
    op.create_index(op.f("ix_operations_job_name"), "operations", ["job_name"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic ###
    op.drop_index(op.f("ix_operations_job_name"), table_name="operations")
    op.drop_table("operations")
    op.drop_index(op.f("ix_leases_worker_name"), table_name="leases")
    op.drop_index(op.f("ix_leases_job_name"), table_name="leases")
    op.drop_table("leases")
    op.drop_index(op.f("ix_jobs_stage"), table_name="jobs")
    op.drop_index(op.f("ix_jobs_priority"), table_name="jobs")
    op.drop_index(op.f("ix_jobs_action_digest"), table_name="jobs")
    op.drop_table("jobs")
    op.drop_index(op.f("ix_index_digest_hash"), table_name="index")
    op.drop_index(op.f("ix_index_accessed_timestamp"), table_name="index")
    op.drop_table("index")
    # ### end Alembic commands ###



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from contextlib import contextmanager
from datetime import timedelta
from typing import Dict, Iterator, Optional

from buildgrid._protos.buildgrid.v2.monitoring_pb2 import MetricRecord
from buildgrid.server.context import try_current_instance, try_current_method, try_current_service
from buildgrid.server.enums import MetricRecordType
from buildgrid.server.monitoring import get_monitoring_bus


def _format_metadata(metadata: Optional[Dict[str, str]] = None) -> Dict[str, str]:
    if metadata is None:
        metadata = {}

    metadata = {key: value for key, value in metadata.items() if value is not None}

    # If an instance name is not specifically set, try to find one from the context.
    if "instanceName" not in metadata:
        if (instance_name := try_current_instance()) is not None:
            metadata["instanceName"] = instance_name

    # If the instance name is set and empty, set it to a default for easy querying.
    if metadata.get("instanceName") == "":
        metadata["instanceName"] = "unnamed"

    # If a service name is not specifically set, try to find one from the context.
    if "serviceName" not in metadata:
        if (service := try_current_service()) is not None:
            metadata["serviceName"] = service

    # If a method name is not specifically set, try to find one from the context.
    if "method" not in metadata:
        if (method := try_current_method()) is not None:
            metadata["method"] = method

    return metadata


def create_counter_record(name: str, count: float, **metadata: str) -> MetricRecord:
    counter_record = MetricRecord()

    counter_record.creation_timestamp.GetCurrentTime()
    counter_record.type = MetricRecordType.COUNTER.value
    counter_record.name = name
    counter_record.count = count
    counter_record.metadata.update(_format_metadata(metadata))

    return counter_record


def create_gauge_record(name: str, value: float, **metadata: str) -> MetricRecord:
    gauge_record = MetricRecord()

    gauge_record.creation_timestamp.GetCurrentTime()
    gauge_record.type = MetricRecordType.GAUGE.value
    gauge_record.name = name
    gauge_record.value = value
    gauge_record.metadata.update(_format_metadata(metadata))

    return gauge_record


def create_timer_record(name: str, duration: timedelta, **metadata: str) -> MetricRecord:
    timer_record = MetricRecord()

    timer_record.creation_timestamp.GetCurrentTime()
    timer_record.type = MetricRecordType.TIMER.value
    timer_record.name = name
    timer_record.duration.FromTimedelta(duration)
    timer_record.metadata.update(_format_metadata(metadata))

    return timer_record


def create_distribution_record(name: str, value: float, **metadata: str) -> MetricRecord:
    dist_record = MetricRecord()

    dist_record.creation_timestamp.GetCurrentTime()
    dist_record.type = MetricRecordType.DISTRIBUTION.value
    dist_record.name = name
    dist_record.count = value
    dist_record.metadata.update(_format_metadata(metadata))

    return dist_record


def publish_counter_metric(name: str, count: float, **metadata: str) -> None:
    record = create_counter_record(name, count, **metadata)
    monitoring_bus = get_monitoring_bus()
    monitoring_bus.send_record_nowait(record)


def publish_gauge_metric(name: str, value: float, **metadata: str) -> None:
    record = create_gauge_record(name, value, **metadata)
    monitoring_bus = get_monitoring_bus()
    monitoring_bus.send_record_nowait(record)


def publish_timer_metric(name: str, duration: timedelta, **metadata: str) -> None:
    record = create_timer_record(name, duration, **metadata)
    monitoring_bus = get_monitoring_bus()
    monitoring_bus.send_record_nowait(record)


def publish_distribution_metric(name: str, count: float, **metadata: str) -> None:
    record = create_distribution_record(name, float(count), **metadata)
    monitoring_bus = get_monitoring_bus()
    monitoring_bus.send_record_nowait(record)


@contextmanager
def timer(metric_name: str, **tags: str) -> Iterator[None]:
    start_time = time.perf_counter()
    error = "None"
    try:
        yield
    except Exception as e:
        error = e.__class__.__name__
        raise
    finally:
        run_time = timedelta(seconds=time.perf_counter() - start_time)
        publish_timer_metric(metric_name, run_time, exceptionType=error, **tags)



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import typing

import boto3
import botocore

if typing.TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client

from buildgrid.server.cleanup.janitor.config import S3Config


def get_s3_client(config: S3Config) -> "S3Client":
    try:
        return boto3.client(
            "s3",
            endpoint_url=config.endpoint,
            aws_access_key_id=config.access_key,
            aws_secret_access_key=config.secret_key,
        )
    except Exception as e:
        raise ValueError("Failed to create an S3 client, check the S3 configuration options.") from e


def check_bucket_versioning(s3_client: "S3Client", bucket: str) -> bool:
    try:
        response = s3_client.get_bucket_versioning(Bucket=bucket)
        return bool(response.get("Status"))
    except botocore.exceptions.ClientError:
        return False



# Copyright (C) 2019, 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Remote Action Cache
===================

Provides an interface to a remote Action Cache. This can be used by other
services (e.g. an Execution service) to communicate with a remote cache.

It provides the same API as any other Action Cache instance backend.

"""


from typing import Optional, Tuple

import grpc

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult, Digest
from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.client.actioncache import ActionCacheClient
from buildgrid.server.client.authentication import ClientCredentials
from buildgrid.server.client.capabilities import CapabilitiesInterface
from buildgrid.server.client.channel import setup_channel
from buildgrid.server.context import current_instance
from buildgrid.server.exceptions import GrpcUninitializedError, NotFoundError


class RemoteActionCache(ActionCacheABC):
    def __init__(
        self,
        remote: str,
        instance_name: Optional[str] = None,
        retries: int = 0,
        max_backoff: int = 64,
        request_timeout: Optional[float] = None,
        channel_options: Optional[Tuple[Tuple[str, str], ...]] = None,
        credentials: Optional[ClientCredentials] = None,
    ) -> None:
        """Initialises a new RemoteActionCache instance.

        Args:
            remote (str): URL of the remote ActionCache service to open a
                channel to.
            instance_name (Optional[str]): The instance name of the remote ActionCache
                service. If none, uses the current instance context.
            channel_options (tuple): Optional tuple of channel options to set
                when opening the gRPC channel to the remote.
            credentials (dict): Optional credentials to use when opening
                the gRPC channel. If unset then an insecure channel will be
                used.

        """
        super().__init__()
        self._remote_instance_name = instance_name
        self._remote = remote
        self._channel_options = channel_options
        if credentials is None:
            credentials = {}
        self._credentials = credentials
        self._channel: Optional[grpc.Channel] = None
        self._allow_updates = None  # type: ignore  # TODO STOP THIS
        self._retries = retries
        self._max_backoff = max_backoff
        self._action_cache: Optional[ActionCacheClient] = None

    def start(self) -> None:
        if self._channel is None:
            self._channel, _ = setup_channel(
                self._remote,
                auth_token=self._credentials.get("auth-token"),
                auth_token_refresh_seconds=self._credentials.get("token-refresh-seconds"),
                client_key=self._credentials.get("tls-client-key"),
                client_cert=self._credentials.get("tls-client-cert"),
                server_cert=self._credentials.get("tls-server-cert"),
            )
        if self._action_cache is None:
            self._action_cache = ActionCacheClient(self._channel, self._retries, self._max_backoff)

    def stop(self) -> None:
        if self._channel:
            self._channel.close()

    @property
    def remote_instance_name(self) -> str:
        if self._remote_instance_name is not None:
            return self._remote_instance_name
        return current_instance()

    @property
    def allow_updates(self) -> bool:
        if self._channel is None:
            raise GrpcUninitializedError("Remote cache used before gRPC initialization.")

        # Check if updates are allowed if we haven't already.
        # This is done the first time update_action_result is called rather
        # than on instantiation because the remote cache may not be running
        # when this object is instantiated.
        if self._allow_updates is None:
            interface = CapabilitiesInterface(self._channel)
            capabilities = interface.get_capabilities(self.remote_instance_name)
            self._allow_updates = capabilities.cache_capabilities.action_cache_update_capabilities.update_enabled
        return self._allow_updates

    def get_action_result(self, action_digest: Digest) -> ActionResult:
        """Retrieves the cached result for an Action.

        Queries the remote ActionCache service to retrieve the cached
        result for a given Action digest. If the remote cache doesn't
        contain a result for the Action, then ``None`` is returned.

        Args:
            action_digest (Digest): The digest of the Action to retrieve the
                cached result of.

        """
        if self._action_cache is None:
            raise GrpcUninitializedError("Remote cache used before gRPC initialization.")

        action_result = self._action_cache.get(self.remote_instance_name, action_digest)

        if action_result is None:
            key = self._get_key(action_digest)
            raise NotFoundError(f"Key not found: {key}")
        return action_result

    def update_action_result(self, action_digest: Digest, action_result: ActionResult) -> None:
        """Stores a result for an Action in the remote cache.

        Sends an ``UpdateActionResult`` request to the remote ActionCache
        service, to store the result in the remote cache.

        If the remote cache doesn't allow updates, then this raises a
        ``NotImplementedError``.

        Args:
            action_digest (Digest): The digest of the Action whose result is
                being cached.
            action_result (ActionResult): The result to cache for the given
                Action digest.

        """
        if self._action_cache is None:
            raise GrpcUninitializedError("Remote cache used before gRPC initialization.")

        if not self.allow_updates:
            raise NotImplementedError("Updating cache not allowed")

        self._action_cache.update(self.remote_instance_name, action_digest, action_result)

    def _get_key(self, action_digest: Digest) -> Tuple[str, int]:
        """Get a hashable cache key from a given Action digest.

        Args:
            action_digest (Digest): The digest to produce a cache key for.

        """
        return (action_digest.hash, action_digest.size_bytes)



# Copyright (C) 2022 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import hashlib
import json

from buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2 import BotSession
from buildgrid._protos.google.devtools.remoteworkers.v1test2.worker_pb2 import Device, Worker
from buildgrid.server.scheduler import DynamicPropertySet
from buildgrid.server.scheduler.properties import hash_from_dict


def session_for_props(props):
    properties = []
    for key, values in props.items():
        if isinstance(values, str):
            values = [values]
        for value in values:
            properties.append(Device.Property(key=key, value=value))

    return BotSession(worker=Worker(devices=[Device(properties=properties)]))


def session_hashes(props, session):
    return list(map(hash_from_dict, props.worker_properties(session)))


def test_get_partial_capabilities_hashes() -> None:
    empty_props = DynamicPropertySet(
        unique_property_keys=set(),
        match_property_keys=set(),
        wildcard_property_keys=set(),
    )
    capabilities = session_hashes(empty_props, BotSession())
    assert sorted(capabilities) == [hashlib.sha1(json.dumps({}, sort_keys=True).encode()).hexdigest()]

    expected_partial_capabilities = [
        {},
        {"OSFamily": ["Linux"]},
        {"ISA": ["x86-32"]},
        {"ISA": ["x86-64"]},
        {"OSFamily": ["Linux"], "ISA": ["x86-32"]},
        {"OSFamily": ["Linux"], "ISA": ["x86-64"]},
        {"ISA": ["x86-32", "x86-64"]},
        {"OSFamily": ["Linux"], "ISA": ["x86-32", "x86-64"]},
    ]
    expected_partial_capabilities_hashes = sorted(
        list(
            map(
                lambda cap: hashlib.sha1(json.dumps(cap, sort_keys=True).encode()).hexdigest(),
                expected_partial_capabilities,
            )
        )
    )
    props = DynamicPropertySet(
        unique_property_keys=set(),
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys=set(),
    )

    capabilities = session_hashes(props, session_for_props({"OSFamily": "Linux", "ISA": {"x86-32", "x86-64"}}))
    assert sorted(capabilities) == expected_partial_capabilities_hashes

    # Should be the same if the string is passed in as a singleton set
    capabilities = session_hashes(props, session_for_props({"OSFamily": {"Linux"}, "ISA": {"x86-32", "x86-64"}}))
    assert sorted(capabilities) == expected_partial_capabilities_hashes

    # Changing the order of the ISA values should produce the same hashes
    capabilities = session_hashes(props, session_for_props({"OSFamily": "Linux", "ISA": {"x86-64", "x86-32"}}))
    assert sorted(capabilities) == expected_partial_capabilities_hashes



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from typing import Dict, List, Literal, Optional, Union

import yaml
from pydantic import BaseModel, Field


class Acl(BaseModel):
    actor: Optional[str] = Field(default=None)
    requests: Optional[List[str]] = Field(default=None)
    subject: Optional[str] = Field(default=None)
    workflow: Optional[str] = Field(default=None)

    def is_authorized(
        self,
        request_name: str,
        actor: Optional[str] = None,
        subject: Optional[str] = None,
        workflow: Optional[str] = None,
    ) -> bool:
        if self.actor is not None and not re.match(self.actor, actor or ""):
            return False

        if self.subject is not None and not re.match(self.subject, subject or ""):
            return False

        if self.workflow is not None and not re.match(self.workflow, workflow or ""):
            return False

        if self.requests is not None and request_name not in self.requests:
            return False

        return True


class InstanceAuthorizationConfig(BaseModel):
    allow: Union[Literal["all"], List[Acl]]

    def is_authorized(
        self,
        request_name: str,
        actor: Optional[str] = None,
        subject: Optional[str] = None,
        workflow: Optional[str] = None,
    ) -> bool:
        if self.allow == "all":
            return True

        return any(acl.is_authorized(request_name, actor, subject, workflow) for acl in self.allow)


def parse_auth_config(path: Union[str, bytes, "os.PathLike[str]"]) -> Dict[str, InstanceAuthorizationConfig]:
    with open(path) as config_file:
        config = yaml.safe_load(config_file)
    return {
        instance_name: InstanceAuthorizationConfig(**instance_config)
        for instance_name, instance_config in config.items()
    }



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/devtools/remoteworkers/v1test2/worker.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2google/devtools/remoteworkers/v1test2/worker.proto\x12%google.devtools.remoteworkers.v1test2\"\xa9\x02\n\x06Worker\x12>\n\x07\x64\x65vices\x18\x01 \x03(\x0b\x32-.google.devtools.remoteworkers.v1test2.Device\x12J\n\nproperties\x18\x02 \x03(\x0b\x32\x36.google.devtools.remoteworkers.v1test2.Worker.Property\x12\x45\n\x07\x63onfigs\x18\x03 \x03(\x0b\x32\x34.google.devtools.remoteworkers.v1test2.Worker.Config\x1a&\n\x08Property\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x1a$\n\x06\x43onfig\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x8c\x01\n\x06\x44\x65vice\x12\x0e\n\x06handle\x18\x01 \x01(\t\x12J\n\nproperties\x18\x02 \x03(\x0b\x32\x36.google.devtools.remoteworkers.v1test2.Device.Property\x1a&\n\x08Property\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\tB\xc3\x01\n)com.google.devtools.remoteworkers.v1test2B\x13RemoteWorkersWorkerP\x01ZRgoogle.golang.org/genproto/googleapis/devtools/remoteworkers/v1test2;remoteworkers\xa2\x02\x02RW\xaa\x02%Google.DevTools.RemoteWorkers.V1Test2b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'buildgrid._protos.google.devtools.remoteworkers.v1test2.worker_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n)com.google.devtools.remoteworkers.v1test2B\023RemoteWorkersWorkerP\001ZRgoogle.golang.org/genproto/googleapis/devtools/remoteworkers/v1test2;remoteworkers\242\002\002RW\252\002%Google.DevTools.RemoteWorkers.V1Test2'
  _globals['_WORKER']._serialized_start=94
  _globals['_WORKER']._serialized_end=391
  _globals['_WORKER_PROPERTY']._serialized_start=315
  _globals['_WORKER_PROPERTY']._serialized_end=353
  _globals['_WORKER_CONFIG']._serialized_start=355
  _globals['_WORKER_CONFIG']._serialized_end=391
  _globals['_DEVICE']._serialized_start=394
  _globals['_DEVICE']._serialized_end=534
  _globals['_DEVICE_PROPERTY']._serialized_start=315
  _globals['_DEVICE_PROPERTY']._serialized_end=353
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Optional

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult, Digest
from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.cas.storage.redis import redis_client_exception_wrapper
from buildgrid.server.cas.storage.storage_abc import StorageABC
from buildgrid.server.enums import ActionCacheEntryType
from buildgrid.server.exceptions import NotFoundError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.redis.provider import RedisProvider

LOGGER = buildgrid_logger(__name__)


class RedisActionCache(ActionCacheABC):
    def __init__(
        self,
        storage: StorageABC,
        redis: RedisProvider,
        allow_updates: bool = True,
        cache_failed_actions: bool = True,
        entry_type: Optional[ActionCacheEntryType] = ActionCacheEntryType.ACTION_RESULT_DIGEST,
        migrate_entries: Optional[bool] = False,
    ) -> None:
        """Initialises a new ActionCache instance using Redis.
        Stores the `ActionResult` message as a value.

            Args:
                storage (StorageABC): storage backend instance to be used to store ActionResults.
                redis (RedisProvider): Redis connection provider
                allow_updates (bool): allow the client to write to storage
                cache_failed_actions (bool): whether to store failed actions in the Action Cache
                entry_type (ActionCacheEntryType): whether to store ActionResults or their digests.
                migrate_entries (bool): if set, migrate entries that contain a value with
                    a different `ActionCacheEntryType` to `entry_type` as they are accessed
                    (False by default).
        """
        super().__init__(storage=storage, allow_updates=allow_updates)

        self._redis = redis
        self._cache_failed_actions = cache_failed_actions
        self._entry_type = entry_type
        self._migrate_entries = migrate_entries

    @redis_client_exception_wrapper
    def get_action_result(self, action_digest: Digest) -> ActionResult:
        key = self._get_key(action_digest)
        action_result = self._get_action_result(key, action_digest)
        if action_result is not None:
            if self.referenced_blobs_still_exist(action_digest, action_result):
                return action_result

            if self._allow_updates:
                LOGGER.debug(
                    "Removing digest from cache due to missing blobs in CAS.", tags=dict(digest=action_digest)
                )
                self._redis.execute_rw(lambda r: r.delete(key))

        raise NotFoundError(f"Key not found: [{key}]")

    @redis_client_exception_wrapper
    def update_action_result(self, action_digest: Digest, action_result: ActionResult) -> None:
        if not self._allow_updates:
            raise NotImplementedError("Updating cache not allowed")

        if self._cache_failed_actions or action_result.exit_code == 0:
            assert self._storage, "Storage used before initialization"
            action_result_digest = self._storage.put_message(action_result)

            cache_key = self._get_key(action_digest)

            if self._entry_type == ActionCacheEntryType.ACTION_RESULT_DIGEST:
                self._redis.execute_rw(lambda r: r.set(cache_key, action_result_digest.SerializeToString()))
            else:
                self._redis.execute_rw(lambda r: r.set(cache_key, action_result.SerializeToString()))

            LOGGER.info("Result cached for action.", tags=dict(digest=action_digest))

    def _get_key(self, action_digest: Digest) -> str:
        return f"action-cache.{action_digest.hash}_{action_digest.size_bytes}"

    def _get_action_result(self, key: str, action_digest: Digest) -> Optional[ActionResult]:
        value_in_cache = self._redis.execute_ro(lambda r: r.get(key))

        if value_in_cache is None:
            return None

        # Attempting to parse the entry as a `Digest` first:
        action_result_digest = Digest.FromString(value_in_cache)
        if len(action_result_digest.hash) == len(action_digest.hash):
            # The cache contains the `Digest` of the `ActionResult`:
            assert self._storage, "Storage used before initialization"
            action_result = self._storage.get_message(action_result_digest, ActionResult)

            # If configured, update the entry to contain an `ActionResult`:
            if self._entry_type == ActionCacheEntryType.ACTION_RESULT and self._migrate_entries:
                LOGGER.debug("Converting entry from Digest to ActionResult.", tags=dict(digest=action_digest))
                assert action_result, "Returned result was none"
                result = action_result.SerializeToString()
                self._redis.execute_rw(lambda r: r.set(key, result))
        else:
            action_result = ActionResult.FromString(value_in_cache)

            # If configured, update the entry to contain a `Digest`:
            if self._entry_type == ActionCacheEntryType.ACTION_RESULT_DIGEST and self._migrate_entries:
                LOGGER.debug("Converting entry from ActionResult to Digest.", tags=dict(digest=action_digest))
                assert self._storage, "Storage used before initialization"
                action_result_digest = self._storage.put_message(action_result)
                self._redis.execute_rw(lambda r: r.set(key, action_result_digest.SerializeToString()))

        return action_result



# Copyright (C) 2021 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
PublishBuildEvent service
=========================

"""


from typing import Iterable, Iterator

import grpc
from google.protobuf.empty_pb2 import Empty

from buildgrid._protos.buildgrid.v2.query_build_events_pb2 import DESCRIPTOR as QBE_DESCRIPTOR
from buildgrid._protos.buildgrid.v2.query_build_events_pb2 import QueryEventStreamsRequest, QueryEventStreamsResponse
from buildgrid._protos.buildgrid.v2.query_build_events_pb2_grpc import (
    QueryBuildEventsServicer,
    add_QueryBuildEventsServicer_to_server,
)
from buildgrid._protos.google.devtools.build.v1.build_events_pb2 import BuildEvent
from buildgrid._protos.google.devtools.build.v1.publish_build_event_pb2 import DESCRIPTOR as PBE_DESCRIPTOR
from buildgrid._protos.google.devtools.build.v1.publish_build_event_pb2 import (
    OrderedBuildEvent,
    PublishBuildToolEventStreamRequest,
    PublishBuildToolEventStreamResponse,
    PublishLifecycleEventRequest,
)
from buildgrid._protos.google.devtools.build.v1.publish_build_event_pb2_grpc import (
    PublishBuildEventServicer,
    add_PublishBuildEventServicer_to_server,
)
from buildgrid.server.build_events.storage import BuildEventStream, BuildEventStreamStorage
from buildgrid.server.decorators import rpc
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.servicer import InstancedServicer

LOGGER = buildgrid_logger(__name__)


def is_lifecycle_event(event: BuildEvent) -> bool:
    lifecycle_events = [
        "build_enqueued",
        "invocation_attempt_started",
        "invocation_attempt_finished",
        "build_finished",
    ]
    return event.WhichOneof("event") in lifecycle_events


class PublishBuildEventService(PublishBuildEventServicer, InstancedServicer[BuildEventStreamStorage]):
    SERVICE_NAME = "PublishBuildEvent"
    REGISTER_METHOD = add_PublishBuildEventServicer_to_server
    FULL_NAME = PBE_DESCRIPTOR.services_by_name[SERVICE_NAME].full_name

    @rpc()
    def PublishLifecycleEvent(self, request: PublishLifecycleEventRequest, context: grpc.ServicerContext) -> Empty:
        """Handler for PublishLifecycleEvent requests.

        This method takes a request containing a build lifecycle event, and
        uses it to update the high-level state of a build (with a corresponding)
        event stream.

        """
        ordered_build_event = request.build_event
        if is_lifecycle_event(ordered_build_event.event):
            stream = self._get_stream_for_event(ordered_build_event)
            stream.publish_event(ordered_build_event)

        else:
            LOGGER.warning("Got a build tool event in a PublishLifecycleEvent request.")
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)

        return Empty()

    @rpc()
    def PublishBuildToolEventStream(
        self,
        request_iterator: Iterable[PublishBuildToolEventStreamRequest],
        context: grpc.ServicerContext,
    ) -> Iterator[PublishBuildToolEventStreamResponse]:
        for request in request_iterator:
            LOGGER.info("Got a BuildToolEvent on the stream.")

            # We don't need to give any special treatment to BuildToolEvents, so
            # just call the underlying `get_stream` method here.
            stream = self.get_instance("").get_stream(request.ordered_build_event.stream_id)

            # `stream` should never be `None`, but in case the internals change
            # in future lets be safe.
            if stream is not None:
                stream.publish_event(request.ordered_build_event)

            yield PublishBuildToolEventStreamResponse(
                stream_id=request.ordered_build_event.stream_id,
                sequence_number=request.ordered_build_event.sequence_number,
            )

    def _get_stream_for_event(self, event: OrderedBuildEvent) -> BuildEventStream:
        # If this is the start of a new build, then we want a new stream
        if event.event.WhichOneof("event") == "build_enqueued":
            return self.get_instance("").new_stream(event.stream_id)
        return self.get_instance("").get_stream(event.stream_id)


class QueryBuildEventsService(QueryBuildEventsServicer, InstancedServicer[BuildEventStreamStorage]):
    SERVICE_NAME = "QueryBuildEvents"
    REGISTER_METHOD = add_QueryBuildEventsServicer_to_server
    FULL_NAME = QBE_DESCRIPTOR.services_by_name[SERVICE_NAME].full_name

    @rpc()
    def QueryEventStreams(
        self, request: QueryEventStreamsRequest, context: grpc.ServicerContext
    ) -> QueryEventStreamsResponse:
        streams = self.get_instance("").get_matching_streams(stream_key_regex=request.build_id_pattern)
        return QueryEventStreamsResponse(streams=[stream.to_grpc_message() for stream in streams])



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CAS Janitor
===========

Cleanup daemon which is responsible for picking up any mess left by the regular cleaner.

The regular cleaner is required to synchronize between the index and the storage, and
so it is possible to get into a state where the blob is removed from the index but not
removed from the storage (eg. in the event of an error when removing the blob from
storage after removing it's index entry).

The Janitor takes care of this situation by regularly iterating over the entire
storage and removing any blobs which aren't referenced by an index. This ensures that
any hanging blobs are removed in a timely manner and the storage space consumption
doesn't spiral up out of control.

"""


import os
import signal
import sys

import click

from buildgrid.server.cleanup.janitor.config import parse_janitor_config
from buildgrid.server.cleanup.janitor.index import IndexLookup, RedisIndexLookup, SqlIndexLookup
from buildgrid.server.cleanup.janitor.s3 import S3Janitor

from ..cli import setup_logging


@click.group(name="janitor", short_help="Service to clean up orphaned blobs from S3 buckets.")
def cli() -> None:
    pass


@cli.command("start", short_help="Start a CAS Janitor process.")
@click.argument("CONFIG_PATH", type=click.Path(file_okay=True, dir_okay=False, exists=True, writable=False))
@click.option("-v", "--verbose", count=True, help="Increase log verbosity level.")
def start(
    config_path: "os.PathLike[str]",
    verbose: int,
) -> None:
    setup_logging(verbosity=verbose)

    click.echo(f"Starting up a CAS janitor: config_path={config_path}")

    config = parse_janitor_config(config_path)

    # Janitoring for an SQL CAS index
    if config.sql_connection_string is not None:
        index: IndexLookup = SqlIndexLookup(config.sql_connection_string)

    # Janitoring for a Redis CAS index
    elif config.redis is not None:
        index = RedisIndexLookup(config.redis)

    # No other janitor types are supported
    else:
        click.echo("Either redis or sql_connection_string is needed", err=True)
        sys.exit(-1)

    janitor = S3Janitor(config, index)
    signal.signal(signal.SIGTERM, janitor.stop)
    signal.signal(signal.SIGINT, janitor.stop)
    try:
        janitor.start()
    finally:
        janitor.stop()



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
ExecutionInstance
=================
An instance of the Remote Execution Service.
"""


from contextlib import ExitStack
from typing import Iterable, Optional

from buildgrid_metering.models.dataclasses import Identity, RPCUsage, Usage

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import DESCRIPTOR as RE_DESCRIPTOR
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import (
    Action,
    Command,
    Digest,
    Platform,
    RequestMetadata,
)
from buildgrid._protos.google.longrunning.operations_pb2 import Operation
from buildgrid.server.auth.manager import get_context_client_identity
from buildgrid.server.enums import MeteringThrottleAction
from buildgrid.server.exceptions import FailedPreconditionError, InvalidArgumentError, ResourceExhaustedError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.scheduler import Scheduler
from buildgrid.server.servicer import Instance
from buildgrid.server.sql.models import ClientIdentityEntry
from buildgrid.server.utils.cancellation import CancellationContext

# All priorities >= this value will not be throttled / deprioritized
EXECUTION_DEPRIORITIZATION_LIMIT = 1

LOGGER = buildgrid_logger(__name__)


class ExecutionInstance(Instance):
    SERVICE_NAME = RE_DESCRIPTOR.services_by_name["Execution"].full_name

    def __init__(
        self,
        scheduler: Scheduler,
        operation_stream_keepalive_timeout: Optional[int] = None,
    ) -> None:
        self._stack = ExitStack()
        self.scheduler = scheduler

        self._operation_stream_keepalive_timeout = operation_stream_keepalive_timeout

    def start(self) -> None:
        self.scheduler.start(start_job_watcher=True)
        self._stack.callback(self.scheduler.stop)

    def stop(self) -> None:
        self._stack.close()
        LOGGER.info("Stopped Execution.")

    def execute(
        self,
        *,
        action_digest: Digest,
        skip_cache_lookup: bool,
        priority: int = 0,
        request_metadata: Optional[RequestMetadata] = None,
        client_identity: Optional[ClientIdentityEntry] = None,
    ) -> str:
        """
        Sends a job for execution. Queues an action and creates an Operation to be associated with this action.
        """

        action = self.scheduler.storage.get_message(action_digest, Action)
        if not action:
            raise FailedPreconditionError("Could not get action from storage.")

        command = self.scheduler.storage.get_message(action.command_digest, Command)
        if not command:
            raise FailedPreconditionError("Could not get command from storage.")

        if action.HasField("platform"):
            platform = action.platform
        elif command.HasField("platform"):
            platform = command.platform
        else:
            platform = Platform()

        property_label, platform_requirements = self.scheduler.property_set.execution_properties(platform)

        should_throttle = self._should_throttle_execution(priority, client_identity)
        if should_throttle:
            if self.scheduler.metering_throttle_action == MeteringThrottleAction.REJECT:
                raise ResourceExhaustedError("User quota exceeded")

            # TODO test_execution_instance is a total mess. It mocks way too much making tests
            # brittle. when possible merge it into execution_service tests and use proper logging here.
            # Should be able to write `action_digest=[{action_digest.hash}/{action_digest.size_bytes}]`, but cant
            # AttributeError: 'str' object has no attribute 'hash'
            LOGGER.info(
                "Job priority throttled.",
                tags=dict(digest=action_digest, old_priority=priority, new_priority=EXECUTION_DEPRIORITIZATION_LIMIT),
            )
            priority = EXECUTION_DEPRIORITIZATION_LIMIT

        operation_name = self.scheduler.queue_job_action(
            action=action,
            action_digest=action_digest,
            command=command,
            platform_requirements=platform_requirements,
            property_label=property_label,
            skip_cache_lookup=skip_cache_lookup,
            priority=priority,
            request_metadata=request_metadata,
            client_identity=client_identity,
        )
        self._meter_execution(client_identity, operation_name)
        return operation_name

    def stream_operation_updates(self, operation_name: str, context: CancellationContext) -> Iterable[Operation]:
        job_name = self.scheduler.get_operation_job_name(operation_name)
        if not job_name:
            raise InvalidArgumentError(f"Operation name does not exist: [{operation_name}]")
        # Start the listener as soon as we get the job name and re-query. This avoids potentially missing
        # the completed update if it triggers in between sending back the first result and the yield resuming.
        with self.scheduler.ops_notifier.subscription(job_name) as update_requested:
            yield (operation := self.scheduler.load_operation(operation_name))
            if operation.done:
                return

            # When the context is deactivated, we can quickly stop waiting.
            context.on_cancel(update_requested.set)
            while not context.is_cancelled():
                update_requested.wait(timeout=self._operation_stream_keepalive_timeout)
                update_requested.clear()

                if context.is_cancelled():
                    return

                yield (operation := self.scheduler.load_operation(operation_name))
                if operation.done:
                    return

    def _meter_execution(self, client_identity: Optional[ClientIdentityEntry], operation_name: str) -> None:
        """Meter the number of executions of client"""
        if self.scheduler.metering_client is None or client_identity is None:
            return
        try:
            identity = Identity(
                instance=client_identity.instance,
                workflow=client_identity.workflow,
                actor=client_identity.actor,
                subject=client_identity.subject,
            )
            usage = Usage(rpc=RPCUsage(execute=1))
            self.scheduler.metering_client.put_usage(identity, operation_name, usage)
        except Exception as exc:
            LOGGER.exception(
                f"Failed to publish execution usage for identity: {get_context_client_identity()}", exc_info=exc
            )

    def _should_throttle_execution(self, priority: int, client_identity: Optional[ClientIdentityEntry]) -> bool:
        if (
            priority >= EXECUTION_DEPRIORITIZATION_LIMIT
            or self.scheduler.metering_client is None
            or client_identity is None
        ):
            return False
        try:
            identity = Identity(
                instance=client_identity.instance,
                workflow=client_identity.workflow,
                actor=client_identity.actor,
                subject=client_identity.subject,
            )
            response = self.scheduler.metering_client.get_throttling(identity)
            if response.throttled:
                LOGGER.info(
                    "Execution request is throttled.",
                    tags=dict(client_id=client_identity, usage=response.tracked_usage),
                )
            return response.throttled
        except Exception as exc:
            LOGGER.exception("Failed to get throttling information.", exc_info=exc)
            return False



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from google.protobuf.timestamp_pb2 import Timestamp

from buildgrid._protos.buildgrid.v2.introspection_pb2 import (
    DESCRIPTOR,
    ListWorkersRequest,
    ListWorkersResponse,
    OperationFilter,
    OperationFilters,
    RegisteredWorker,
    WorkerOperation,
)
from buildgrid.server.enums import BotStatus
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.operations.filtering.interpreter import VALID_OPERATION_FILTERS, OperationFilterSpec
from buildgrid.server.operations.filtering.sanitizer import DatetimeValueSanitizer, SortKeyValueSanitizer
from buildgrid.server.scheduler.impl import Scheduler
from buildgrid.server.servicer import Instance
from buildgrid.server.settings import MAX_LIST_PAGE_SIZE
from buildgrid.server.utils.digests import parse_digest


class IntrospectionInstance(Instance):
    SERVICE_NAME = DESCRIPTOR.services_by_name["Introspection"].full_name

    def __init__(self, sql: Scheduler) -> None:
        self._scheduler = sql

    def list_workers(self, request: ListWorkersRequest) -> ListWorkersResponse:
        if request.page_size < 0:
            raise InvalidArgumentError("Page size must be a positive integer")
        if request.page < 0:
            raise InvalidArgumentError("Page number must be a positive integer")

        page = request.page or 1
        page_size = min(request.page_size, MAX_LIST_PAGE_SIZE) or MAX_LIST_PAGE_SIZE
        bot_entries, count = self._scheduler.list_workers(request.worker_name, page, page_size)

        workers = []
        for bot in bot_entries:
            last_update = Timestamp()
            last_update.FromDatetime(bot.last_update_timestamp)
            expiry: Optional[Timestamp] = None
            if bot.expiry_time is not None:
                expiry = Timestamp()
                expiry.FromDatetime(bot.expiry_time)
            worker = RegisteredWorker(
                worker_name=bot.bot_id,
                session_name=bot.name,
                last_updated=last_update,
                bot_status=BotStatus(bot.bot_status).value,
                expiry_time=expiry,
            )
            if bot.job is not None:
                if digest := parse_digest(bot.job.action_digest):
                    worker.action_digest.CopyFrom(digest)
                worker.operations.extend(
                    [WorkerOperation(operation_name=operation.name) for operation in bot.job.operations]
                )
            workers.append(worker)

        return ListWorkersResponse(workers=workers, total=count, page=page, page_size=page_size)

    def get_operation_filters(self) -> OperationFilters:
        def _generate_filter_spec(key: str, spec: OperationFilterSpec) -> OperationFilter:
            comparators = ["<", "<=", "=", "!=", ">=", ">"]
            filter_type = "text"
            if isinstance(spec.sanitizer, SortKeyValueSanitizer):
                comparators = ["="]
            elif isinstance(spec.sanitizer, DatetimeValueSanitizer):
                filter_type = "datetime"

            try:
                values = spec.sanitizer.valid_values
            except NotImplementedError:
                values = []

            return OperationFilter(
                key=key,
                name=spec.name,
                type=filter_type,
                description=spec.description,
                comparators=comparators,
                values=values,
            )

        return OperationFilters(
            filters=[_generate_filter_spec(key, spec) for key, spec in VALID_OPERATION_FILTERS.items()]
        )



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Union

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import DESCRIPTOR as RE_DESCRIPTOR
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import (
    ActionCacheUpdateCapabilities,
    CacheCapabilities,
    ExecutionCapabilities,
    ServerCapabilities,
    SymlinkAbsolutePathStrategy,
)
from buildgrid._protos.build.bazel.semver.semver_pb2 import SemVer
from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.actioncache.instance import ActionCache
from buildgrid.server.cas.instance import ContentAddressableStorageInstance
from buildgrid.server.execution.instance import ExecutionInstance
from buildgrid.server.servicer import Instance
from buildgrid.server.settings import HIGH_REAPI_VERSION, LOW_REAPI_VERSION, MAX_REQUEST_SIZE
from buildgrid.server.utils.digests import get_hash_type

ActionCacheInstance = Union[ActionCache, ActionCacheABC]


class CapabilitiesInstance(Instance):
    SERVICE_NAME = RE_DESCRIPTOR.services_by_name["Capabilities"].full_name

    def __init__(
        self,
        cas_instance: Optional["ContentAddressableStorageInstance"] = None,
        action_cache_instance: Optional["ActionCacheInstance"] = None,
        execution_instance: Optional["ExecutionInstance"] = None,
    ) -> None:
        self._cas_instance = cas_instance
        self._action_cache_instance = action_cache_instance
        self._execution_instance = execution_instance
        self._high_api_version = _split_semantic_version(HIGH_REAPI_VERSION)
        self._low_api_version = _split_semantic_version(LOW_REAPI_VERSION)

    def add_cas_instance(self, cas_instance: "ContentAddressableStorageInstance") -> None:
        self._cas_instance = cas_instance

    def add_action_cache_instance(self, action_cache_instance: "ActionCacheInstance") -> None:
        self._action_cache_instance = action_cache_instance

    def add_execution_instance(self, execution_instance: "ExecutionInstance") -> None:
        self._execution_instance = execution_instance

    def get_capabilities(self) -> ServerCapabilities:
        # The caching capabilities are VERY hardcoded by the buildgrid server components.
        # Simply forwarding abilities from remote storage or local components does more
        # harm than good, as the remote capabilities do not feed into usage by most components.
        #
        # The most sane way to override these values would be via server configuration,
        # not by dynamically fetching the capabilities remotely.
        #
        # If we want to properly support this functionality, a large refactor would be needed.
        return ServerCapabilities(
            cache_capabilities=CacheCapabilities(
                digest_functions=[get_hash_type()],
                max_batch_total_size_bytes=MAX_REQUEST_SIZE,
                symlink_absolute_path_strategy=SymlinkAbsolutePathStrategy.DISALLOWED,
                action_cache_update_capabilities=ActionCacheUpdateCapabilities(
                    update_enabled=bool(self._action_cache_instance and self._action_cache_instance.allow_updates)
                ),
            ),
            execution_capabilities=ExecutionCapabilities(
                exec_enabled=self._execution_instance is not None,
                digest_function=get_hash_type(),
                digest_functions=[get_hash_type()],
            ),
            low_api_version=self._low_api_version,
            high_api_version=self._high_api_version,
        )


def _split_semantic_version(version_string: str) -> SemVer:
    major_version, minor_version, patch_version = version_string.split(".")

    semantic_version = SemVer()
    semantic_version.major = int(major_version)
    semantic_version.minor = int(minor_version)
    semantic_version.patch = int(patch_version)

    return semantic_version



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Storage Instances
=================
Instances of CAS and ByteStream
"""

from datetime import timedelta
from typing import Iterable, Iterator, List, Optional, Sequence, Set, Tuple

from cachetools import TTLCache
from grpc import RpcError

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import DESCRIPTOR as RE_DESCRIPTOR
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import (
    BatchReadBlobsResponse,
    BatchUpdateBlobsRequest,
    BatchUpdateBlobsResponse,
    Digest,
    Directory,
    FindMissingBlobsResponse,
    GetTreeRequest,
    GetTreeResponse,
    Tree,
)
from buildgrid._protos.google.bytestream import bytestream_pb2 as bs_pb2
from buildgrid._protos.google.rpc import code_pb2, status_pb2
from buildgrid.server.cas.storage.storage_abc import StorageABC, create_write_session
from buildgrid.server.exceptions import (
    IncompleteReadError,
    InvalidArgumentError,
    NotFoundError,
    OutOfRangeError,
    PermissionDeniedError,
    RetriableError,
)
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric, publish_distribution_metric
from buildgrid.server.servicer import Instance
from buildgrid.server.settings import HASH, HASH_LENGTH, MAX_REQUEST_COUNT, MAX_REQUEST_SIZE, STREAM_ERROR_RETRY_PERIOD
from buildgrid.server.utils.digests import create_digest

LOGGER = buildgrid_logger(__name__)

EMPTY_BLOB = b""
EMPTY_BLOB_DIGEST: Digest = create_digest(EMPTY_BLOB)


class ContentAddressableStorageInstance(Instance):
    SERVICE_NAME = RE_DESCRIPTOR.services_by_name["ContentAddressableStorage"].full_name

    def __init__(
        self,
        storage: StorageABC,
        read_only: bool = False,
        tree_cache_size: Optional[int] = None,
        tree_cache_ttl_minutes: float = 60,
    ) -> None:
        self._storage = storage
        self.__read_only = read_only

        self._tree_cache: Optional[TTLCache[Tuple[str, int], Digest]] = None
        if tree_cache_size:
            self._tree_cache = TTLCache(tree_cache_size, tree_cache_ttl_minutes * 60)

    def start(self) -> None:
        self._storage.start()

    def stop(self) -> None:
        self._storage.stop()
        LOGGER.info("Stopped CAS.")

    def find_missing_blobs(self, blob_digests: Sequence[Digest]) -> FindMissingBlobsResponse:
        deduplicated_digests: List[Digest] = []
        seen: Set[str] = set()
        for digest in blob_digests:
            if digest.hash in seen:
                continue
            seen.add(digest.hash)
            deduplicated_digests.append(digest)
        blob_digests = deduplicated_digests

        missing_blobs = self._storage.missing_blobs(blob_digests)

        num_blobs_in_request = len(blob_digests)
        if num_blobs_in_request > 0:
            num_blobs_missing = len(missing_blobs)
            percent_missing = float((num_blobs_missing / num_blobs_in_request) * 100)

            publish_distribution_metric(METRIC.CAS.BLOBS_COUNT, num_blobs_in_request)
            publish_distribution_metric(METRIC.CAS.BLOBS_MISSING_COUNT, num_blobs_missing)
            publish_distribution_metric(METRIC.CAS.BLOBS_MISSING_PERCENT, percent_missing)

        for digest in blob_digests:
            publish_distribution_metric(METRIC.CAS.BLOB_BYTES, digest.size_bytes)

        return FindMissingBlobsResponse(missing_blob_digests=missing_blobs)

    def batch_update_blobs(self, requests: Sequence[BatchUpdateBlobsRequest.Request]) -> BatchUpdateBlobsResponse:
        if self.__read_only:
            raise PermissionDeniedError("CAS is read-only")

        if len(requests) > 0:
            publish_distribution_metric(METRIC.CAS.BLOBS_COUNT, len(requests))

        storage = self._storage
        store = []
        seen: Set[str] = set()
        for request_proto in requests:
            if request_proto.digest.hash in seen:
                continue
            seen.add(request_proto.digest.hash)
            store.append((request_proto.digest, request_proto.data))
            publish_distribution_metric(METRIC.CAS.BLOB_BYTES, request_proto.digest.size_bytes)

        response = BatchUpdateBlobsResponse()
        statuses = storage.bulk_update_blobs(store)

        for (digest, _), status in zip(store, statuses):
            response_proto = response.responses.add()
            response_proto.digest.CopyFrom(digest)
            response_proto.status.CopyFrom(status)

        return response

    def batch_read_blobs(self, digests: Sequence[Digest]) -> BatchReadBlobsResponse:
        storage = self._storage

        if len(digests) > 0:
            publish_distribution_metric(METRIC.CAS.BLOBS_COUNT, len(digests))

        # Only process unique digests
        good_digests = []
        bad_digests = []
        seen: Set[str] = set()
        requested_bytes = 0
        for digest in digests:
            if digest.hash in seen:
                continue
            seen.add(digest.hash)

            if len(digest.hash) != HASH_LENGTH:
                bad_digests.append(digest)
            else:
                good_digests.append(digest)
                requested_bytes += digest.size_bytes

        if requested_bytes > MAX_REQUEST_SIZE:
            raise InvalidArgumentError(
                "Combined total size of blobs exceeds "
                "server limit. "
                f"({requested_bytes} > {MAX_REQUEST_SIZE} [byte])"
            )

        if len(good_digests) > 0:
            blobs_read = storage.bulk_read_blobs(good_digests)
        else:
            blobs_read = {}

        response = BatchReadBlobsResponse()

        for digest in good_digests:
            response_proto = response.responses.add()
            response_proto.digest.CopyFrom(digest)

            if digest.hash in blobs_read and blobs_read[digest.hash] is not None:
                response_proto.data = blobs_read[digest.hash]
                status_code = code_pb2.OK

                publish_distribution_metric(METRIC.CAS.BLOB_BYTES, digest.size_bytes)
            else:
                status_code = code_pb2.NOT_FOUND
                LOGGER.info("Blob not found from BatchReadBlobs.", tags=dict(digest=digest))

            response_proto.status.CopyFrom(status_pb2.Status(code=status_code))

        for digest in bad_digests:
            response_proto = response.responses.add()
            response_proto.digest.CopyFrom(digest)
            status_code = code_pb2.INVALID_ARGUMENT
            response_proto.status.CopyFrom(status_pb2.Status(code=status_code))

        return response

    def lookup_tree_cache(self, root_digest: Digest) -> Optional[Tree]:
        """Find full Tree from cache"""
        if self._tree_cache is None:
            return None
        tree = None
        if response_digest := self._tree_cache.get((root_digest.hash, root_digest.size_bytes)):
            tree = self._storage.get_message(response_digest, Tree)
            if tree is None:
                self._tree_cache.pop((root_digest.hash, root_digest.size_bytes))

        publish_counter_metric(METRIC.CAS.TREE_CACHE_HIT_COUNT, 1 if tree else 0)
        return tree

    def put_tree_cache(self, root_digest: Digest, root: Directory, children: Iterable[Directory]) -> None:
        """Put Tree with a full list of directories into CAS"""
        if self._tree_cache is None:
            return
        tree = Tree(root=root, children=children)
        tree_digest = self._storage.put_message(tree)
        self._tree_cache[(root_digest.hash, root_digest.size_bytes)] = tree_digest

    def get_tree(self, request: GetTreeRequest) -> Iterator[GetTreeResponse]:
        storage = self._storage

        if not request.page_size:
            request.page_size = MAX_REQUEST_COUNT

        if tree := self.lookup_tree_cache(request.root_digest):
            # Cache hit, yield responses based on page size
            directories = [tree.root]
            directories.extend(tree.children)
            yield from (
                GetTreeResponse(directories=directories[start : start + request.page_size])
                for start in range(0, len(directories), request.page_size)
            )
            return

        results = []
        response = GetTreeResponse()

        for dir in storage.get_tree(request.root_digest):
            response.directories.append(dir)
            results.append(dir)
            if len(response.directories) >= request.page_size:
                yield response
                response.Clear()

        if response.directories:
            yield response
        if results:
            self.put_tree_cache(request.root_digest, results[0], results[1:])


class ByteStreamInstance(Instance):
    SERVICE_NAME = bs_pb2.DESCRIPTOR.services_by_name["ByteStream"].full_name

    BLOCK_SIZE = 1 * 1024 * 1024  # 1 MB block size

    def __init__(
        self,
        storage: StorageABC,
        read_only: bool = False,
        disable_overwrite_early_return: bool = False,
    ) -> None:
        self._storage = storage
        self._query_activity_timeout = 30

        self.__read_only = read_only

        # If set, prevents `ByteStream.Write()` from returning without
        # reading all the client's `WriteRequests` for a digest that is
        # already in storage (i.e. not follow the REAPI-specified
        # behavior).
        self.__disable_overwrite_early_return = disable_overwrite_early_return
        # (Should only be used to work around issues with implementations
        # that treat the server half-closing its end of the gRPC stream
        # as a HTTP/2 stream error.)

    def start(self) -> None:
        self._storage.start()

    def stop(self) -> None:
        self._storage.stop()
        LOGGER.info("Stopped ByteStream.")

    def read_cas_blob(self, digest: Digest, read_offset: int, read_limit: int) -> Iterator[bs_pb2.ReadResponse]:
        digest_str = f"'{digest.hash}/{digest.size_bytes}'"
        # Check the given read offset and limit.
        if read_offset < 0 or read_offset > digest.size_bytes:
            raise OutOfRangeError(f"Read offset out of range for {digest_str}: {read_offset=}")

        if read_limit < 0:
            raise InvalidArgumentError(f"Read limit out of range for {digest_str}: {read_limit=}")

        bytes_requested = digest.size_bytes - read_offset
        if read_limit:
            bytes_requested = min(read_limit, bytes_requested)

        if bytes_requested == 0:
            yield bs_pb2.ReadResponse(data=b"")
            return

        bytes_remaining = bytes_requested

        # Read the blob from storage and send its contents to the client.
        result = self._storage.get_blob(digest)
        if result is None:
            raise NotFoundError(f"Blob not found for {digest_str}")

        try:
            if read_offset > 0:
                result.seek(read_offset)

            publish_distribution_metric(METRIC.CAS.BLOB_BYTES, digest.size_bytes)

            # https://docs.python.org/3/library/io.html#io.RawIOBase.read
            # If 0 bytes are returned, and size was not 0, this indicates end of file.
            while block_data := result.read(min(self.BLOCK_SIZE, bytes_remaining)):
                bytes_remaining -= len(block_data)
                yield bs_pb2.ReadResponse(data=block_data)
        finally:
            result.close()

        if bytes_remaining != 0:
            raise IncompleteReadError(
                f"Blob incomplete: {digest_str}, from Bytestream.Read. "
                f"Only read {bytes_requested - bytes_remaining} bytes out of "
                f"requested {bytes_requested} bytes. {read_offset=} {read_limit=}"
            )

    def write_cas_blob(
        self, digest_hash: str, digest_size: str, requests: Iterator[bs_pb2.WriteRequest]
    ) -> bs_pb2.WriteResponse:
        if self.__read_only:
            raise PermissionDeniedError("ByteStream is read-only")

        if len(digest_hash) != HASH_LENGTH or not digest_size.isdigit():
            raise InvalidArgumentError(f"Invalid digest [{digest_hash}/{digest_size}]")

        digest = Digest(hash=digest_hash, size_bytes=int(digest_size))

        publish_distribution_metric(METRIC.CAS.BLOB_BYTES, digest.size_bytes)

        if self._storage.has_blob(digest):
            # According to the REAPI specification:
            # "When attempting an upload, if another client has already
            # completed the upload (which may occur in the middle of a single
            # upload if another client uploads the same blob concurrently),
            # the request will terminate immediately [...]".
            #
            # However, half-closing the stream can be problematic with some
            # intermediaries like HAProxy.
            # (https://github.com/haproxy/haproxy/issues/1219)
            #
            # If half-closing the stream is not allowed, we read and drop
            # all the client's messages before returning, still saving
            # the cost of a write to storage.
            if self.__disable_overwrite_early_return:
                try:
                    for request in requests:
                        if request.finish_write:
                            break
                        continue
                except RpcError:
                    msg = "ByteStream client disconnected whilst streaming requests, upload cancelled."
                    LOGGER.debug(msg)
                    raise RetriableError(msg, retry_period=timedelta(seconds=STREAM_ERROR_RETRY_PERIOD))

            return bs_pb2.WriteResponse(committed_size=digest.size_bytes)

        # Start the write session and write the first request's data.

        with create_write_session(digest) as write_session:
            computed_hash = HASH()

            # Handle subsequent write requests.
            bytes_count = 0
            try:
                for request in requests:
                    write_session.write(request.data)

                    computed_hash.update(request.data)
                    bytes_count += len(request.data)

                    if request.finish_write:
                        break
            except RpcError:
                write_session.close()
                msg = "ByteStream client disconnected whilst streaming requests, upload cancelled."
                LOGGER.debug(msg)
                raise RetriableError(msg, retry_period=timedelta(seconds=STREAM_ERROR_RETRY_PERIOD))

            # Check that the data matches the provided digest.
            if bytes_count != digest.size_bytes:
                raise NotImplementedError(
                    "Cannot close stream before finishing write, "
                    f"got {bytes_count} bytes but expected {digest.size_bytes}"
                )

            if computed_hash.hexdigest() != digest.hash:
                raise InvalidArgumentError("Data does not match hash")

            self._storage.commit_write(digest, write_session)
            return bs_pb2.WriteResponse(committed_size=bytes_count)



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class AuthMetadataMethod(Enum):
    # No authorization:
    NONE = "none"
    # Authorize based on client identity headers:
    HEADERS = "headers"
    # JWT based authorization:
    JWT = "jwt"


class AuthMetadataAlgorithm(Enum):
    # No encryption involved:
    UNSPECIFIED = "unspecified"
    # JWT related algorithms:
    JWT_ES256 = "es256"  # ECDSA signature algorithm using SHA-256 hash algorithm
    JWT_ES384 = "es384"  # ECDSA signature algorithm using SHA-384 hash algorithm
    JWT_ES512 = "es512"  # ECDSA signature algorithm using SHA-512 hash algorithm
    JWT_HS256 = "hs256"  # HMAC using SHA-256 hash algorithm
    JWT_HS384 = "hs384"  # HMAC using SHA-384 hash algorithm
    JWT_HS512 = "hs512"  # HMAC using SHA-512 hash algorithm
    JWT_PS256 = "ps256"  # RSASSA-PSS using SHA-256 and MGF1 padding with SHA-256
    JWT_PS384 = "ps384"  # RSASSA-PSS signature using SHA-384 and MGF1 padding with SHA-384
    JWT_PS512 = "ps512"  # RSASSA-PSS signature using SHA-512 and MGF1 padding with SHA-512
    JWT_RS256 = "rs256"  # RSASSA-PKCS1-v1_5 signature algorithm using SHA-256 hash algorithm
    JWT_RS384 = "rs384"  # RSASSA-PKCS1-v1_5 signature algorithm using SHA-384 hash algorithm
    JWT_RS512 = "rs512"  # RSASSA-PKCS1-v1_5 signature algorithm using SHA-512 hash algorithm



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from buildgrid._protos.buildgrid.v2 import introspection_pb2 as buildgrid_dot_v2_dot_introspection__pb2


class IntrospectionStub(object):
    """BuildGrid's Introspection service is used to query the state of
    BuildGrid's internals, and obtain details on available functionality.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ListWorkers = channel.unary_unary(
                '/buildgrid.v2.Introspection/ListWorkers',
                request_serializer=buildgrid_dot_v2_dot_introspection__pb2.ListWorkersRequest.SerializeToString,
                response_deserializer=buildgrid_dot_v2_dot_introspection__pb2.ListWorkersResponse.FromString,
                )
        self.GetOperationFilters = channel.unary_unary(
                '/buildgrid.v2.Introspection/GetOperationFilters',
                request_serializer=buildgrid_dot_v2_dot_introspection__pb2.GetOperationFiltersRequest.SerializeToString,
                response_deserializer=buildgrid_dot_v2_dot_introspection__pb2.OperationFilters.FromString,
                )


class IntrospectionServicer(object):
    """BuildGrid's Introspection service is used to query the state of
    BuildGrid's internals, and obtain details on available functionality.
    """

    def ListWorkers(self, request, context):
        """Query the list of workers known to a specific BuildGrid instance

        Errors:

        * `INVALID_ARGUMENT`: When the provided instance name doesn't exist
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetOperationFilters(self, request, context):
        """Get the list of valid filters for use in ListOperations requests
        for a specific BuildGrid instance

        Errors:

        * `INVALID_ARGUMENT`: When the provided instance name doesn't exist
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_IntrospectionServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ListWorkers': grpc.unary_unary_rpc_method_handler(
                    servicer.ListWorkers,
                    request_deserializer=buildgrid_dot_v2_dot_introspection__pb2.ListWorkersRequest.FromString,
                    response_serializer=buildgrid_dot_v2_dot_introspection__pb2.ListWorkersResponse.SerializeToString,
            ),
            'GetOperationFilters': grpc.unary_unary_rpc_method_handler(
                    servicer.GetOperationFilters,
                    request_deserializer=buildgrid_dot_v2_dot_introspection__pb2.GetOperationFiltersRequest.FromString,
                    response_serializer=buildgrid_dot_v2_dot_introspection__pb2.OperationFilters.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'buildgrid.v2.Introspection', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Introspection(object):
    """BuildGrid's Introspection service is used to query the state of
    BuildGrid's internals, and obtain details on available functionality.
    """

    @staticmethod
    def ListWorkers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/buildgrid.v2.Introspection/ListWorkers',
            buildgrid_dot_v2_dot_introspection__pb2.ListWorkersRequest.SerializeToString,
            buildgrid_dot_v2_dot_introspection__pb2.ListWorkersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetOperationFilters(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/buildgrid.v2.Introspection/GetOperationFilters',
            buildgrid_dot_v2_dot_introspection__pb2.GetOperationFiltersRequest.SerializeToString,
            buildgrid_dot_v2_dot_introspection__pb2.OperationFilters.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: buildgrid/v2/introspection.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2 as build_dot_bazel_dot_remote_dot_execution_dot_v2_dot_remote__execution__pb2
from buildgrid._protos.google.devtools.remoteworkers.v1test2 import bots_pb2 as google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_bots__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n buildgrid/v2/introspection.proto\x12\x0c\x62uildgrid.v2\x1a\x36\x62uild/bazel/remote/execution/v2/remote_execution.proto\x1a\x30google/devtools/remoteworkers/v1test2/bots.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"a\n\x12ListWorkersRequest\x12\x15\n\rinstance_name\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x0c\n\x04page\x18\x03 \x01(\x05\x12\x13\n\x0bworker_name\x18\x04 \x01(\t\")\n\x0fWorkerOperation\x12\x16\n\x0eoperation_name\x18\x01 \x01(\t\"\xd9\x02\n\x10RegisteredWorker\x12\x14\n\x0csession_name\x18\x01 \x01(\t\x12\x13\n\x0bworker_name\x18\x02 \x01(\t\x12\x30\n\x0clast_updated\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x44\n\nbot_status\x18\x04 \x01(\x0e\x32\x30.google.devtools.remoteworkers.v1test2.BotStatus\x12/\n\x0b\x65xpiry_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\noperations\x18\x06 \x03(\x0b\x32\x1d.buildgrid.v2.WorkerOperation\x12>\n\raction_digest\x18\x07 \x01(\x0b\x32\'.build.bazel.remote.execution.v2.Digest\"v\n\x13ListWorkersResponse\x12\x0c\n\x04page\x18\x01 \x01(\x05\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\r\n\x05total\x18\x03 \x01(\x05\x12/\n\x07workers\x18\x04 \x03(\x0b\x32\x1e.buildgrid.v2.RegisteredWorker\"3\n\x1aGetOperationFiltersRequest\x12\x15\n\rinstance_name\x18\x01 \x01(\t\"t\n\x0fOperationFilter\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x13\n\x0b\x63omparators\x18\x05 \x03(\t\x12\x0e\n\x06values\x18\x06 \x03(\t\"B\n\x10OperationFilters\x12.\n\x07\x66ilters\x18\x01 \x03(\x0b\x32\x1d.buildgrid.v2.OperationFilter2\xc8\x01\n\rIntrospection\x12T\n\x0bListWorkers\x12 .buildgrid.v2.ListWorkersRequest\x1a!.buildgrid.v2.ListWorkersResponse\"\x00\x12\x61\n\x13GetOperationFilters\x12(.buildgrid.v2.GetOperationFiltersRequest\x1a\x1e.buildgrid.v2.OperationFilters\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'buildgrid._protos.buildgrid.v2.introspection_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_LISTWORKERSREQUEST']._serialized_start=189
  _globals['_LISTWORKERSREQUEST']._serialized_end=286
  _globals['_WORKEROPERATION']._serialized_start=288
  _globals['_WORKEROPERATION']._serialized_end=329
  _globals['_REGISTEREDWORKER']._serialized_start=332
  _globals['_REGISTEREDWORKER']._serialized_end=677
  _globals['_LISTWORKERSRESPONSE']._serialized_start=679
  _globals['_LISTWORKERSRESPONSE']._serialized_end=797
  _globals['_GETOPERATIONFILTERSREQUEST']._serialized_start=799
  _globals['_GETOPERATIONFILTERSREQUEST']._serialized_end=850
  _globals['_OPERATIONFILTER']._serialized_start=852
  _globals['_OPERATIONFILTER']._serialized_end=968
  _globals['_OPERATIONFILTERS']._serialized_start=970
  _globals['_OPERATIONFILTERS']._serialized_end=1036
  _globals['_INTROSPECTION']._serialized_start=1039
  _globals['_INTROSPECTION']._serialized_end=1239
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2021-2022 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
import time
from typing import Any, Callable, TypeVar

import grpc

from buildgrid.server.exceptions import (
    CancelledError,
    FailedPreconditionError,
    InvalidArgumentError,
    NotFoundError,
    PermissionDeniedError,
    StorageFullError,
)

T = TypeVar("T")


class GrpcRetrier:
    def __init__(self, retries: int, max_backoff: int = 64, should_backoff: bool = True):
        """Initializes a new :class:`GrpcRetrier`.

        Args:
            retries (int): The maximum number of attempts for each RPC call.
            max_backoff (int): The maximum time to wait between retries.
            should_backoff (bool): Whether to backoff at all. Always set this to True except in tests.
        """

        self._retries = retries
        self._max_backoff = max_backoff
        self._should_backoff = should_backoff

    def retry(self, func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
        attempts = 0
        while True:
            try:
                return func(*args, **kwargs)
            except grpc.RpcError as e:
                status_code = e.code()

                # Retry only on UNAVAILABLE and ABORTED
                if status_code in (grpc.StatusCode.UNAVAILABLE, grpc.StatusCode.ABORTED):
                    attempts += 1
                    if attempts > self._retries:
                        raise ConnectionError(e.details()) from e
                    if self._should_backoff:
                        # Sleep for 2^(N-1) + jitter seconds, where N is # of attempts
                        jitter = random.uniform(0, 1)
                        time.sleep(min(pow(2, attempts - 1) + jitter, self._max_backoff))

                elif status_code == grpc.StatusCode.CANCELLED:
                    raise CancelledError(e.details()) from e

                elif status_code == grpc.StatusCode.INVALID_ARGUMENT:
                    raise InvalidArgumentError(e.details()) from e

                elif status_code == grpc.StatusCode.DEADLINE_EXCEEDED:
                    raise TimeoutError(e.details()) from e

                elif status_code == grpc.StatusCode.NOT_FOUND:
                    raise NotFoundError("Requested data does not exist on remote") from e

                elif status_code == grpc.StatusCode.PERMISSION_DENIED:
                    raise PermissionDeniedError(e.details()) from e

                elif status_code == grpc.StatusCode.RESOURCE_EXHAUSTED:
                    raise StorageFullError(e.details()) from e

                elif status_code == grpc.StatusCode.FAILED_PRECONDITION:
                    raise FailedPreconditionError(e.details()) from e

                elif status_code == grpc.StatusCode.UNIMPLEMENTED:
                    raise NotImplementedError(e.details()) from e

                else:
                    raise



# Copyright (C) 2022 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from contextvars import ContextVar
from typing import Any, Dict, Iterable, Optional, Tuple, Union, cast

from grpc.aio import Metadata

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import RequestMetadata, ToolDetails
from buildgrid._protos.buildgrid.v2.identity_pb2 import ClientIdentity
from buildgrid.server.auth.manager import get_context_client_identity
from buildgrid.server.settings import (
    CLIENT_IDENTITY_HEADER_NAME,
    REQUEST_METADATA_HEADER_NAME,
    REQUEST_METADATA_TOOL_NAME,
    REQUEST_METADATA_TOOL_VERSION,
)
from buildgrid.server.sql.models import ClientIdentityEntry

ctx_request_metadata: ContextVar[RequestMetadata] = ContextVar(
    "ctx_request_metadata",
    default=RequestMetadata(
        tool_details=ToolDetails(
            tool_name=REQUEST_METADATA_TOOL_NAME,
            tool_version=REQUEST_METADATA_TOOL_VERSION,
        ),
    ),
)


def metadata_list() -> Tuple[Tuple[str, Union[str, bytes]], ...]:
    """Helper function to construct the metadata list from the ContextVar."""
    metadata = ctx_request_metadata.get()
    return ((REQUEST_METADATA_HEADER_NAME, metadata.SerializeToString()),)


ctx_grpc_request_id: ContextVar[Optional[str]] = ContextVar("grpc_request_id", default=None)


def printable_request_metadata(metadata_entries: Any) -> str:
    """Given a metadata object, return a human-readable representation
    of its `RequestMetadata` entry.

    Args:
        metadata_entries: tuple of entries obtained from a gRPC context
            with, for example, `context.invocation_metadata()`.

    Returns:
        A string with the metadata contents.
    """
    metadata = extract_request_metadata(metadata_entries)
    return request_metadata_to_string(metadata)


def extract_request_metadata_dict(metadata_entries: Any) -> Dict[str, str]:
    metadata = extract_request_metadata(metadata_entries)
    return request_metadata_to_dict(metadata)


def extract_request_metadata(metadata_entries: Any) -> RequestMetadata:
    """Given a list of string tuples, extract the RequestMetadata
    header values if they are present. If they were not provided,
    returns an empty message.

    Args:
        metadata_entries: tuple of entries obtained from a gRPC context
            with, for example, `context.invocation_metadata()`.

    Returns:
        A `RequestMetadata` proto. If the metadata is not defined in the
        request, the message will be empty.
    """
    request_metadata_entry = next(
        (entry for entry in metadata_entries if entry.key == REQUEST_METADATA_HEADER_NAME), None
    )

    request_metadata = RequestMetadata()
    if request_metadata_entry:
        request_metadata.ParseFromString(request_metadata_entry.value)
    return request_metadata


def request_metadata_to_string(request_metadata: RequestMetadata) -> str:
    if request_metadata.tool_details:
        tool_name = request_metadata.tool_details.tool_name
        tool_version = request_metadata.tool_details.tool_version
    else:
        tool_name = tool_version = ""

    return (
        f'tool_name="{tool_name}", tool_version="{tool_version}", '
        f'action_id="{request_metadata.action_id}", '
        f'tool_invocation_id="{request_metadata.tool_invocation_id}", '
        f'correlated_invocations_id="{request_metadata.correlated_invocations_id}", '
        f'action_mnemonic="{request_metadata.action_mnemonic}", '
        f'target_id="{request_metadata.target_id}", '
        f'configuration_id="{request_metadata.configuration_id}"'
    )


def request_metadata_to_dict(request_metadata: RequestMetadata) -> Dict[str, str]:
    if request_metadata.tool_details:
        tool_name = request_metadata.tool_details.tool_name
        tool_version = request_metadata.tool_details.tool_version
    else:
        tool_name = tool_version = ""

    return {
        "tool_name": tool_name,
        "tool_version": tool_version,
        "action_id": request_metadata.action_id,
        "tool_invocation_id": request_metadata.tool_invocation_id,
        "correlated_invocations_id": request_metadata.correlated_invocations_id,
        "action_mnemonic": request_metadata.action_mnemonic,
        "target_id": request_metadata.target_id,
        "configuration_id": request_metadata.configuration_id,
    }


def extract_client_identity_dict(instance: str, invocation_metadata: Iterable[Tuple[str, Any]]) -> Dict[str, Any]:
    client_id = extract_client_identity(instance, invocation_metadata)
    return {
        "client_id": client_id.id if client_id else "",
        "instance": client_id.instance if client_id else "",
        "workflow": client_id.workflow if client_id else "",
        "actor": client_id.actor if client_id else "",
        "subject": client_id.subject if client_id else "",
    }


def extract_client_identity(
    instance: str, invocation_metadata: Iterable[Tuple[str, Any]]
) -> Optional[ClientIdentityEntry]:
    """Checks for the existence of the client identity in the ClientIdentity
    context var. If the context var is not set then extracts
    the ClientIdentity from gRPC metadata

    Args:
        instance (str): the instance where the request was invoked from
        invocation_metadata (List[Tuple[str, str]]): grpc metadata

    Returns:
        Optional[ClientIdentityEntry]: identity of the client if exists
    """
    context_client_identity = get_context_client_identity()
    if (
        context_client_identity
        and context_client_identity.actor
        and context_client_identity.subject
        and context_client_identity.workflow
    ):
        return ClientIdentityEntry(
            instance=instance,
            workflow=context_client_identity.workflow,
            actor=context_client_identity.actor,
            subject=context_client_identity.subject,
        )

    metadata_dict = dict(invocation_metadata)
    workflow = metadata_dict.get("x-request-workflow", None)
    actor = metadata_dict.get("x-request-actor", None)
    subject = metadata_dict.get("x-request-subject", None)

    # None or empty strings are invalid
    if workflow and actor and subject:
        return ClientIdentityEntry(instance=instance, workflow=workflow, actor=actor, subject=subject)

    return None


def printable_client_identity(instance: str, invocation_metadata: Iterable[Tuple[str, Any]]) -> str:
    """Given a metadata object, return a human-readable representation
    of its `ClientIdentity` entry.

    Args:
        instance: REAPI instance name
        invocation_metadata: tuple of entries obtained from a gRPC context
            with, for example, `context.invocation_metadata()`.

    Returns:
        A string with the ClientIdentity contents.
    """
    client_id = extract_client_identity(instance, invocation_metadata)
    return str(client_id)


def extract_trailing_client_identity(metadata_entries: Metadata) -> ClientIdentity:
    """Given a list of string tuples, extract the ClientIdentity
    header values if they are present. If they were not provided,
    returns an empty message.

    Args:
        metadata_entries: Sequence of gRPC trailing metadata headers.

    Returns:
        A `ClientIdentity` proto. If the metadata is not defined in the
        request, the message will be empty.
    """
    client_identity_entry = next(
        (entry for entry in metadata_entries if entry[0] == CLIENT_IDENTITY_HEADER_NAME), None
    )

    client_identity = ClientIdentity()
    if client_identity_entry:
        client_identity.ParseFromString(cast(bytes, client_identity_entry[1]))
    return client_identity



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add cascades for job_platforms table

Revision ID: bc324dfd3610
Revises: afa46425330d
Create Date: 2023-12-01 11:30:54.380837

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "bc324dfd3610"
down_revision = "afa46425330d"
branch_labels = None
depends_on = None

naming_convention = {
    "fk": "job_platforms_job_name_fkey",
}


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_context().bind
    assert bind is not None, "Alembic database connection is missing"

    with op.batch_alter_table("job_platforms", naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint("job_platforms_job_name_fkey", type_="foreignkey")
        batch_op.create_foreign_key(
            "job_platforms_job_name_fkey", "jobs", ["job_name"], ["name"], onupdate="CASCADE", ondelete="CASCADE"
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    bind = op.get_context().bind
    assert bind is not None, "Alembic database connection is missing"

    with op.batch_alter_table("job_platforms", naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint("job_platforms_job_name_fkey", type_="foreignkey")
        batch_op.create_foreign_key("job_platforms_job_name_fkey", "jobs", ["job_name"], ["name"])



# Copyright (C) 2021 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

from buildgrid.server.utils.async_lru_cache import LruCache


def test_get_missing_key():
    async def _test():
        cache = LruCache(10)
        assert await cache.get("foo") is None
        return True

    assert asyncio.run(_test())


def test_set_and_get():
    async def _test():
        cache: LruCache[str] = LruCache(10)
        key = "foo"
        value = "foo value"

        await cache.set(key, value, 10)

        assert await cache.size() == 1
        assert await cache.get(key) == value
        return True

    assert asyncio.run(_test())


def test_expiry():
    async def _test():
        cache: LruCache[str] = LruCache(2)
        await cache.set("a", "vala", 10)
        await cache.set("b", "valb", 10)

        # Add a third key, taking us over the max size
        await cache.set("c", "valc", 10)

        # Ensure that the oldest key was dropped
        assert await cache.get("a") is None
        assert await cache.get("b") == "valb"
        assert await cache.get("c") == "valc"
        return True

    assert asyncio.run(_test())


def test_ttl_expiry():
    async def _test():
        cache: LruCache[str] = LruCache(10)

        # Set a value with a short ttl
        await cache.set("key", "value", 1)

        # Wait for longer than the ttl
        await asyncio.sleep(1.1)

        # Check that the value isn't returned
        assert await cache.get("key") is None
        assert await cache.size() == 0
        return True

    assert asyncio.run(_test())


def test_zero_ttl():
    async def _test():
        cache: LruCache[str] = LruCache(1)

        # Set a value with a zero ttl (ie. live forever)
        await cache.set("key", "value", 0)

        # Wait for a second
        await asyncio.sleep(1)

        # Check that the value hasn't expired
        assert await cache.get("key") == "value"

        # Set a new key, to force expiry via LRU
        await cache.set("new", "value", 10)
        assert await cache.get("key") is None
        return True

    assert asyncio.run(_test())


def test_set():
    async def _test():
        cache: LruCache[str] = LruCache(2)
        await cache.set("a", "vala", 10)
        await cache.set("b", "valb", 10)

        # Check that setting updates the value
        await cache.set("a", "value a", 10)
        assert await cache.get("a") == "value a"

        # Check that setting updates the ttl
        await cache.set("c", "valc", 1)
        await cache.set("c", "valc", 10)
        await asyncio.sleep(1.1)
        assert await cache.get("c") == "valc"

        # Check that expiry is actually LRU
        assert await cache.get("b") is None
        return True

    assert asyncio.run(_test())


def test_get_updates_lru():
    async def _test():
        cache: LruCache[str] = LruCache(2)
        await cache.set("a", "vala", 10)
        await cache.set("b", "valb", 10)

        # Touch the first key we set
        await cache.get("a")

        # Add a new key, taking us over the max size
        await cache.set("c", "valc", 10)

        # Ensure that the second key we set was dropped
        assert await cache.get("b") is None
        assert await cache.get("a") == "vala"
        return True

    assert asyncio.run(_test())



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import abc
import buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2
import grpc

class TasksStub:
    """DEPRECATED. GetTask should be replaced by Lease.payload, UpdateTaskResult by
    Lease.result and logs should be precreated prior to sending to the bot (eg,
    via CommandTask.expected_outputs.stdout_destination).
    """

    def __init__(self, channel: grpc.Channel) -> None: ...
    GetTask: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.GetTaskRequest,
        buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.Task,
    ]
    """DEPRECATED - use Lease.payload instead.
    GetTask reads the current state of the task. Tasks must be created through
    some other interface, and should be immutable once created and exposed to
    the bots.
    """
    UpdateTaskResult: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.UpdateTaskResultRequest,
        buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.TaskResult,
    ]
    """DEPRECATED - use Lease.result instead.
    UpdateTaskResult updates the result.
    """
    AddTaskLog: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.AddTaskLogRequest,
        buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.AddTaskLogResponse,
    ]
    """DEPRECATED - precreate logs prior to sending to bot.
    AddTaskLog creates a new streaming log. The log is streamed and marked as
    completed through other interfaces (i.e., ByteStream). This can be called
    by the bot if it wants to create a new log; the server can also predefine
    logs that do not need to be created (e.g. `stdout`).
    """

class TasksServicer(metaclass=abc.ABCMeta):
    """DEPRECATED. GetTask should be replaced by Lease.payload, UpdateTaskResult by
    Lease.result and logs should be precreated prior to sending to the bot (eg,
    via CommandTask.expected_outputs.stdout_destination).
    """

    @abc.abstractmethod
    def GetTask(
        self,
        request: buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.GetTaskRequest,
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.Task:
        """DEPRECATED - use Lease.payload instead.
        GetTask reads the current state of the task. Tasks must be created through
        some other interface, and should be immutable once created and exposed to
        the bots.
        """
    @abc.abstractmethod
    def UpdateTaskResult(
        self,
        request: buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.UpdateTaskResultRequest,
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.TaskResult:
        """DEPRECATED - use Lease.result instead.
        UpdateTaskResult updates the result.
        """
    @abc.abstractmethod
    def AddTaskLog(
        self,
        request: buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.AddTaskLogRequest,
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.google.devtools.remoteworkers.v1test2.tasks_pb2.AddTaskLogResponse:
        """DEPRECATED - precreate logs prior to sending to bot.
        AddTaskLog creates a new streaming log. The log is streamed and marked as
        completed through other interfaces (i.e., ByteStream). This can be called
        by the bot if it wants to create a new log; the server can also predefine
        logs that do not need to be created (e.g. `stdout`).
        """

def add_TasksServicer_to_server(servicer: TasksServicer, server: grpc.Server) -> None: ...



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import select
import uuid
from contextlib import contextmanager
from threading import Event, Lock
from typing import Any, Dict, Iterator, Tuple, cast

from sqlalchemy import select as sql_select
from sqlalchemy.orm import Session

from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.sql.models import JobEntry
from buildgrid.server.sql.provider import SqlProvider
from buildgrid.server.threading import ContextWorker

LOGGER = buildgrid_logger(__name__)


class OperationsNotifier:
    def __init__(self, sql_provider: SqlProvider, poll_interval: float = 1) -> None:
        """
        Creates a notifier for changes to jobs, used by observes of related operations.

        Note: jobs have a one-to-many relationship with operations, and for each operation
        there can be multiple clients listening for updates.
        """

        self._sql = sql_provider
        self._lock = Lock()
        self._listeners: Dict[str, Dict[str, Event]] = {}
        self.poll_interval = poll_interval
        self.worker = ContextWorker(name="OperationsNotifier", target=self.begin)

    def __enter__(self: "OperationsNotifier") -> "OperationsNotifier":
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.stop()

    def start(self) -> None:
        self.worker.start()

    def stop(self) -> None:
        self.worker.stop()

    def listener_count(self) -> int:
        """Used for reporting job metrics about the scheduling."""
        with self._lock:
            return sum(len(events) for events in self._listeners.values())

    def begin(self, shutdown_requested: Event) -> None:
        LOGGER.info("Starting job notifier thread.")

        while not shutdown_requested.is_set():
            try:
                with self._sql.session() as session:
                    if self._sql.dialect == "postgresql":
                        self._listen_for_updates(shutdown_requested, session)
                    else:
                        self._poll_for_updates(shutdown_requested, session)
            except Exception as e:
                LOGGER.warning(
                    f"OperationsNotifier encountered exception: {e}.",
                    tags=dict(retry_delay_seconds=self.poll_interval),
                )
                # Sleep for a bit so that we give enough time for the
                # database to potentially recover
                shutdown_requested.wait(timeout=self.poll_interval)

    def _listen_for_updates(self, shutdown_requested: Event, session: Session) -> None:
        # In our `LISTEN` call, we want to *bypass the ORM* and *use the underlying Engine connection directly*.
        # This is because using a `session.execute()` will implicitly create a SQL transaction, causing
        # notifications to only be delivered when that transaction is committed.
        from psycopg2.extensions import connection

        try:
            pool_connection = session.connection().connection
            pool_connection.cursor().execute("LISTEN job_updated;")
            pool_connection.commit()
        except Exception:
            LOGGER.warning("Could not start listening to DB for job updates.", exc_info=True)
            raise

        while not shutdown_requested.is_set() and pool_connection.dbapi_connection is not None:
            # If we're in this method, we know we have a psycopg2 connection object here.
            dbapi_connection = cast(connection, pool_connection.dbapi_connection)

            # Wait until the connection is ready for reading. Timeout and try again if there was nothing to read.
            # If the connection becomes readable, collect the notifications it has received and handle them.
            #
            # See https://www.psycopg.org/docs/advanced.html#asynchronous-notifications
            if select.select([dbapi_connection], [], [], self.poll_interval) == ([], [], []):
                continue

            dbapi_connection.poll()
            while dbapi_connection.notifies:
                notify = dbapi_connection.notifies.pop()
                self.notify(notify.payload)

    def _poll_for_updates(self, shutdown_requested: Event, session: Session) -> None:
        prev_data: Dict[str, Tuple[bool, int]] = {}
        while not shutdown_requested.is_set():
            with self._lock:
                names = list(self._listeners)

            # Only query for the minimal amount of data required.
            # The subscribers can choose how they want to act (e.g. by querying the full job data).
            statement = sql_select(JobEntry.name, JobEntry.cancelled, JobEntry.stage).where(JobEntry.name.in_(names))
            next_data: Dict[str, Tuple[bool, int]] = {}
            for [name, cancelled, stage] in session.execute(statement).all():
                next_data[name] = (cancelled, stage)

            for name in next_data:
                if name not in prev_data or prev_data[name] != next_data[name]:
                    self.notify(name)

            prev_data = next_data
            shutdown_requested.wait(timeout=self.poll_interval)

    def notify(self, job_name: str) -> None:
        with self._lock:
            if job_name in self._listeners:
                for event in self._listeners[job_name].values():
                    event.set()

    @contextmanager
    def subscription(self, job_name: str) -> Iterator[Event]:
        """
        Register a threading.Event object which is triggered each time the associated job_name updates
        its cancelled or stage status. After waiting for an event, the caller should immediately call
        event.clear() if they wish to re-use the event again, otherwise the event object will remain set.
        """

        # Create a unique key for the subscription which is deleted when the job is no longer monitored.
        key = str(uuid.uuid4())
        event = Event()
        try:
            with self._lock:
                if job_name not in self._listeners:
                    self._listeners[job_name] = {}
                self._listeners[job_name][key] = event
            yield event
        finally:
            with self._lock:
                del self._listeners[job_name][key]
                if len(self._listeners[job_name]) == 0:
                    del self._listeners[job_name]
